"""Unit tests for kgb.agency.SpyAgency."""

from __future__ import unicode_literals

from contextlib import contextmanager

import kgb.asserts
from kgb.agency import SpyAgency
from kgb.signature import FunctionSig
from kgb.tests.base import MathClass, TestCase


class SpyAgencyTests(TestCase):
    """Unit tests for kgb.agency.SpyAgency."""

    def test_spy_on(self):
        """Testing SpyAgency.spy_on"""
        obj = MathClass()

        spy = self.agency.spy_on(obj.do_math)
        self.assertEqual(self.agency.spies, set([spy]))

    def test_spy_for(self):
        """Testing SpyAgency.spy_for"""
        obj = MathClass()

        @self.agency.spy_for(obj.do_math, owner=obj)
        def my_func(_self, a=None, b=None, *args, **kwargs):
            """Some docs."""
            return 123

        self.assertEqual(obj.do_math(), 123)
        self.assertEqual(my_func(obj), 123)
        self.assertIs(obj.do_math.spy.func, my_func)

        # Make sure we decorated correctly.
        sig = FunctionSig(my_func)
        self.assertEqual(sig.func_name, 'my_func')
        self.assertEqual(sig.func_type, FunctionSig.TYPE_FUNCTION)
        self.assertEqual(my_func.__doc__, 'Some docs.')

    def test_unspy(self):
        """Testing SpyAgency.unspy"""
        obj = MathClass()
        orig_do_math = obj.do_math

        spy = self.agency.spy_on(obj.do_math)
        self.assertEqual(self.agency.spies, set([spy]))
        self.assertTrue(hasattr(obj.do_math, 'spy'))

        self.agency.unspy(obj.do_math)
        self.assertEqual(self.agency.spies, set())
        self.assertFalse(hasattr(obj.do_math, 'spy'))

        self.assertEqual(obj.do_math, orig_do_math)

    def test_unspy_all(self):
        """Testing SpyAgency.unspy_all"""
        obj = MathClass()
        orig_do_math = obj.do_math

        spy1 = self.agency.spy_on(obj.do_math)
        spy2 = self.agency.spy_on(MathClass.class_do_math)

        self.assertEqual(self.agency.spies, set([spy1, spy2]))
        self.assertTrue(hasattr(obj.do_math, 'spy'))
        self.assertTrue(hasattr(MathClass.class_do_math, 'spy'))

        self.agency.unspy_all()
        self.assertEqual(self.agency.spies, set())

        self.assertEqual(obj.do_math, orig_do_math)
        self.assertEqual(MathClass.class_do_math, self.orig_class_do_math)
        self.assertFalse(hasattr(obj.do_math, 'spy'))
        self.assertFalse(hasattr(MathClass.class_do_math, 'spy'))


class TestCaseMixinTests(SpyAgency, TestCase):
    """Unit tests for SpyAgency as a TestCase mixin."""

    def test_spy_on(self):
        """Testing SpyAgency mixed in with spy_on"""
        obj = MathClass()

        self.spy_on(obj.do_math)
        self.assertTrue(hasattr(obj.do_math, 'spy'))

        result = obj.do_math()
        self.assertEqual(result, 3)

    def test_spy_for(self):
        """Testing SpyAgency mixed in with spy_for"""
        obj = MathClass()

        @self.spy_for(obj.do_math)
        def my_func(_self, a=None, b=None, *args, **kwargs):
            """Some docs."""
            return 123

        self.assertEqual(obj.do_math(), 123)
        self.assertEqual(my_func(obj), 123)
        self.assertIs(obj.do_math.spy.func, my_func)

        # Make sure we decorated correctly.
        sig = FunctionSig(my_func)
        self.assertEqual(sig.func_name, 'my_func')
        self.assertEqual(sig.func_type, FunctionSig.TYPE_FUNCTION)
        self.assertEqual(my_func.__doc__, 'Some docs.')

    def test_tear_down(self):
        """Testing SpyAgency mixed in with tearDown"""
        obj = MathClass()
        orig_do_math = obj.do_math
        func_dict = obj.do_math.__dict__.copy()

        self.spy_on(obj.do_math)
        self.assertTrue(hasattr(obj.do_math, 'spy'))
        self.assertNotEqual(func_dict, obj.do_math.__dict__)

        self.tearDown()

        self.assertEqual(obj.do_math, orig_do_math)
        self.assertFalse(hasattr(obj.do_math, 'spy'))
        self.assertEqual(func_dict, obj.do_math.__dict__)

    def test_assertHasSpy_with_spy(self):
        """Testing SpyAgency.assertHasSpy with spy"""
        self.spy_on(MathClass.do_math,
                    owner=MathClass)

        # These should not fail.
        self.assertHasSpy(MathClass.do_math)
        self.assertHasSpy(MathClass.do_math.spy)

        # Check the aliases.
        self.assert_has_spy(MathClass.do_math)
        kgb.asserts.assert_has_spy(MathClass.do_math)

    def test_assertHasSpy_without_spy(self):
        """Testing SpyAgency.assertHasSpy without spy"""
        with self._check_assertion('do_math has not been spied on.'):
            self.assertHasSpy(MathClass.do_math)

    def test_assertSpyCalled_with_called(self):
        """Testing SpyAgency.assertSpyCalled with spy called"""
        obj = MathClass()
        self.spy_on(obj.do_math)

        obj.do_math()

        # These should not fail.
        self.assertSpyCalled(obj.do_math)
        self.assertSpyCalled(obj.do_math.spy)

        # Check the aliases.
        self.assert_spy_called(obj.do_math)
        kgb.asserts.assert_spy_called(obj.do_math)

    def test_assertSpyCalled_without_called(self):
        """Testing SpyAgency.assertSpyCalled without spy called"""
        obj = MathClass()
        self.spy_on(obj.do_math)

        msg = 'do_math was not called.'

        with self._check_assertion(msg):
            self.assertSpyCalled(obj.do_math)

        with self._check_assertion(msg):
            self.assertSpyCalled(obj.do_math.spy)

    def test_assertSpyNotCalled_without_called(self):
        """Testing SpyAgency.assertSpyNotCalled without spy called"""
        obj = MathClass()
        self.spy_on(obj.do_math)

        # These should not fail.
        self.assertSpyNotCalled(obj.do_math)
        self.assertSpyNotCalled(obj.do_math.spy)

        # Check the aliases.
        self.assert_spy_not_called(obj.do_math)
        kgb.asserts.assert_spy_not_called(obj.do_math)

    def test_assertSpyNotCalled_with_called(self):
        """Testing SpyAgency.assertSpyNotCalled with spy called"""
        obj = MathClass()
        self.spy_on(obj.do_math)

        obj.do_math(3, b=4)
        obj.do_math(2, b=9)

        msg = (
            "do_math was called 2 times:\n"
            "\n"
            "Call 0:\n"
            "  args=()\n"
            "  kwargs={'a': 3, 'b': 4}\n"
            "\n"
            "Call 1:\n"
            "  args=()\n"
            "  kwargs={'a': 2, 'b': 9}"
        )

        with self._check_assertion(msg):
            self.assertSpyNotCalled(obj.do_math)

        with self._check_assertion(msg):
            self.assertSpyNotCalled(obj.do_math.spy)

    def test_assertSpyCallCount_with_expected_count(self):
        """Testing SpyAgency.assertSpyCallCount with expected call count"""
        obj = MathClass()
        self.spy_on(obj.do_math)

        obj.do_math()
        obj.do_math()

        # These should not fail.
        self.assertSpyCallCount(obj.do_math, 2)
        self.assertSpyCallCount(obj.do_math.spy, 2)

        # Check the aliases.
        self.assert_spy_call_count(obj.do_math, 2)
        kgb.asserts.assert_spy_call_count(obj.do_math, 2)

    def test_assertSpyCallCount_without_expected_count(self):
        """Testing SpyAgency.assertSpyCallCount without expected call count"""
        obj = MathClass()
        self.spy_on(obj.do_math)

        obj.do_math()

        with self._check_assertion('do_math was called 1 time, not 2.'):
            self.assertSpyCallCount(obj.do_math, 2)

        # Let's bump and test a plural result.
        obj.do_math()

        with self._check_assertion('do_math was called 2 times, not 3.'):
            self.assertSpyCallCount(obj.do_math.spy, 3)

    def test_assertSpyCalledOnce_with_0_calls(self):
        """Testing SpyAgency.assertSpyCalledOnce with 0 calls"""
        obj = MathClass()
        self.spy_on(obj.do_math)

        # This should fail.
        msg = 'do_math was not called.'

        with self._check_assertion(msg):
            self.assertSpyCalledOnce(obj.do_math)

    def test_assertSpyCalledOnce_with_1_calls(self):
        """Testing SpyAgency.assertSpyCalledOnce with 1 call"""
        obj = MathClass()
        self.spy_on(obj.do_math)

        obj.do_math(1, b=4)

        # These should not fail.
        self.assertSpyCalledOnce(obj.do_math)
        self.assertSpyCalledOnce(obj.do_math.spy)

        # Check the aliases.
        self.assert_spy_called_once(obj.do_math)
        kgb.asserts.assert_spy_called_once(obj.do_math)

    def test_assertSpyCalledOnce_with_2_calls(self):
        """Testing SpyAgency.assertSpyCalledOnce with 2 calls"""
        obj = MathClass()
        self.spy_on(obj.do_math)

        obj.do_math(1, b=4)
        obj.do_math(2, b=8)

        # This should fail.
        msg = (
            "do_math was not called exactly 1 time. It was called 2 times:\n"
            "\n"
            "Call 0:\n"
            "  args=()\n"
            "  kwargs={'a': 1, 'b': 4}\n"
            "\n"
            "Call 1:\n"
            "  args=()\n"
            "  kwargs={'a': 2, 'b': 8}"
        )

        with self._check_assertion(msg):
            self.assertSpyCalledOnce(obj.do_math)

    def test_assertSpyCalledOnceWith_with_0_calls(self):
        """Testing SpyAgency.assertSpyCalledOnceWith with 0 calls"""
        obj = MathClass()
        self.spy_on(obj.do_math)

        # This should fail.
        msg = 'do_math was not called.'

        with self._check_assertion(msg):
            self.assertSpyCalledOnceWith(obj.do_math, a=1, b=2)

    def test_assertSpyCalledOnceWith_with_1_call(self):
        """Testing SpyAgency.assertSpyCalledOnceWith with 1 call"""
        obj = MathClass()
        self.spy_on(obj.do_math)

        obj.do_math(1, b=4)

        # This should not fail.
        self.assertSpyCalledOnceWith(obj.do_math, a=1, b=4)
        self.assertSpyCalledOnceWith(obj.do_math.spy, a=1, b=4)

        # Check the aliases.
        self.assert_spy_called_once_with(obj.do_math, a=1, b=4)
        kgb.asserts.assert_spy_called_once_with(obj.do_math, a=1, b=4)

    def test_assertSpyCalledOnceWith_with_1_calls_with_wrong_args(self):
        """Testing SpyAgency.assertSpyCalledOnceWith with 1 call with wrong
        arguments
        """
        obj = MathClass()
        self.spy_on(obj.do_math)

        obj.do_math(1, b=4)

        # This should fail.
        msg = (
            "No call to do_math was passed args=(1,), kwargs={'b': 8}.\n"
            "\n"
            "The following calls were recorded:\n"
            "\n"
            "Call 0:\n"
            "  args=()\n"
            "  kwargs={'a': 1, 'b': 4}"
        )

        with self._check_assertion(msg):
            self.assertSpyCalledOnceWith(obj.do_math, 1, b=8)

    def test_assertSpyCalledOnceWith_with_2_calls(self):
        """Testing SpyAgency.assertSpyCalledOnceWith with 2 calls"""
        obj = MathClass()
        self.spy_on(obj.do_math)

        obj.do_math(1, b=4)
        obj.do_math(2, b=8)

        # This should fail.
        msg = (
            "do_math was not called exactly 1 time. It was called 2 times:\n"
            "\n"
            "Call 0:\n"
            "  args=()\n"
            "  kwargs={'a': 1, 'b': 4}\n"
            "\n"
            "Call 1:\n"
            "  args=()\n"
            "  kwargs={'a': 2, 'b': 8}"
        )

        with self._check_assertion(msg):
            self.assertSpyCalledOnceWith(obj.do_math, 1, b=4)

    def test_assertSpyCalledWith_with_expected_arguments(self):
        """Testing SpyAgency.assertSpyCalledWith with expected arguments"""
        obj = MathClass()
        self.spy_on(obj.do_math)

        obj.do_math(1, b=4)
        obj.do_math(2, b=9)

        # These should not fail.
        self.assertSpyCalledWith(obj.do_math, a=1, b=4)
        self.assertSpyCalledWith(obj.do_math.calls[0], a=1, b=4)
        self.assertSpyCalledWith(obj.do_math.spy, a=2, b=9)
        self.assertSpyCalledWith(obj.do_math.spy.calls[1], a=2, b=9)

        # Check the aliases.
        self.assert_spy_called_with(obj.do_math, a=1, b=4)
        kgb.asserts.assert_spy_called_with(obj.do_math, a=1, b=4)

    def test_assertSpyCalledWith_without_expected_arguments(self):
        """Testing SpyAgency.assertSpyCalledWith without expected arguments"""
        obj = MathClass()
        self.spy_on(obj.do_math)

        obj.do_math(1, b=4)
        obj.do_math(2, b=9)

        msg = (
            "No call to do_math was passed args=(), kwargs={'x': 4, 'z': 1}.\n"
            "\n"
            "The following calls were recorded:\n"
            "\n"
            "Call 0:\n"
            "  args=()\n"
            "  kwargs={'a': 1, 'b': 4}\n"
            "\n"
            "Call 1:\n"
            "  args=()\n"
            "  kwargs={'a': 2, 'b': 9}"
        )

        with self._check_assertion(msg):
            self.assertSpyCalledWith(obj.do_math, x=4, z=1)

        with self._check_assertion(msg):
            self.assertSpyCalledWith(obj.do_math.spy, x=4, z=1)

        msg = (
            "This call to do_math was not passed args=(),"
            " kwargs={'x': 4, 'z': 1}.\n"
            "\n"
            "It was called with:\n"
            "\n"
            "args=()\n"
            "kwargs={'a': 1, 'b': 4}"
        )

        with self._check_assertion(msg):
            self.assertSpyCalledWith(obj.do_math.spy.calls[0], x=4, z=1)

    def test_assertSpyNotCalledWith_with_unexpected_arguments(self):
        """Testing SpyAgency.assertSpyNotCalledWith with unexpected arguments
        """
        obj = MathClass()
        self.spy_on(obj.do_math)

        obj.do_math(1, b=4)
        obj.do_math(2, b=9)

        # These should not fail.
        self.assertSpyNotCalledWith(obj.do_math, a=1, b=3)
        self.assertSpyNotCalledWith(obj.do_math.calls[0], a=1, b=3)
        self.assertSpyNotCalledWith(obj.do_math.spy, a=1, b=9)
        self.assertSpyNotCalledWith(obj.do_math.spy.calls[1], a=1, b=9)

        # Check the aliases.
        self.assert_spy_not_called_with(obj.do_math, a=1, b=3)
        kgb.asserts.assert_spy_not_called_with(obj.do_math, a=1, b=3)

    def test_assertSpyNotCalledWith_without_unexpected_arguments(self):
        """Testing SpyAgency.assertSpyNotCalledWith without unexpected
        arguments
        """
        obj = MathClass()
        self.spy_on(obj.do_math)

        obj.do_math(1, b=4)
        obj.do_math(2, b=9)

        msg = (
            "A call to do_math was unexpectedly passed args=(), "
            "kwargs={'a': 1, 'b': 4}.\n"
            "\n"
            "The following calls were recorded:\n"
            "\n"
            "Call 0:\n"
            "  args=()\n"
            "  kwargs={'a': 1, 'b': 4}\n"
            "\n"
            "Call 1:\n"
            "  args=()\n"
            "  kwargs={'a': 2, 'b': 9}"
        )

        with self._check_assertion(msg):
            self.assertSpyNotCalledWith(obj.do_math, a=1, b=4)

        with self._check_assertion(msg):
            self.assertSpyNotCalledWith(obj.do_math.spy, a=1, b=4)

        msg = (
            "This call to do_math was unexpectedly passed args=(),"
            " kwargs={'a': 2, 'b': 9}."
        )

        with self._check_assertion(msg):
            self.assertSpyNotCalledWith(obj.do_math.spy.calls[1], a=2, b=9)

    def test_assertSpyLastCalledWith_with_expected_arguments(self):
        """Testing SpyAgency.assertSpyLastCalledWith with expected arguments"""
        obj = MathClass()
        self.spy_on(obj.do_math)

        obj.do_math(1, b=4)
        obj.do_math(2, b=9)

        # These should not fail.
        self.assertSpyLastCalledWith(obj.do_math, a=2, b=9)
        self.assertSpyLastCalledWith(obj.do_math.spy, a=2, b=9)

        # Check the aliases.
        self.assert_spy_last_called_with(obj.do_math, a=2, b=9)
        kgb.asserts.assert_spy_last_called_with(obj.do_math, a=2, b=9)

    def test_assertSpyLastCalledWith_without_expected_arguments(self):
        """Testing SpyAgency.assertSpyLastCalledWith without expected
        arguments
        """
        obj = MathClass()
        self.spy_on(obj.do_math)

        obj.do_math(1, b=4)
        obj.do_math(2, b=9)

        msg = (
            "The last call to do_math was not passed args=(),"
            " kwargs={'a': 1, 'b': 4}.\n"
            "\n"
            "It was last called with:\n"
            "\n"
            "args=()\n"
            "kwargs={'a': 2, 'b': 9}"
        )

        with self._check_assertion(msg):
            self.assertSpyLastCalledWith(obj.do_math, a=1, b=4)

        with self._check_assertion(msg):
            self.assertSpyLastCalledWith(obj.do_math.spy, a=1, b=4)

    def test_assertSpyReturned_with_expected_return(self):
        """Testing SpyAgency.assertSpyReturned with expected return value"""
        obj = MathClass()
        self.spy_on(obj.do_math)

        obj.do_math(1, b=4)
        obj.do_math(2, b=9)

        # These should not fail.
        self.assertSpyReturned(obj.do_math, 5)
        self.assertSpyReturned(obj.do_math.calls[0], 5)
        self.assertSpyReturned(obj.do_math.spy, 11)
        self.assertSpyReturned(obj.do_math.spy.calls[1], 11)

        # Check the aliases.
        self.assert_spy_returned(obj.do_math, 5)
        kgb.asserts.assert_spy_returned(obj.do_math, 5)

    def test_assertSpyReturned_without_expected_return(self):
        """Testing SpyAgency.assertSpyReturned without expected return value"""
        obj = MathClass()
        self.spy_on(obj.do_math)

        obj.do_math(1, b=4)
        obj.do_math(2, b=9)

        msg = (
            'No call to do_math returned 100.\n'
            '\n'
            'The following values have been returned:\n'
            '\n'
            'Call 0:\n'
            '  5\n'
            '\n'
            'Call 1:\n'
            '  11'
        )

        with self._check_assertion(msg):
            self.assertSpyReturned(obj.do_math, 100)

        with self._check_assertion(msg):
            self.assertSpyReturned(obj.do_math.spy, 100)

        msg = (
            'This call to do_math did not return 100.\n'
            '\n'
            'It returned:\n'
            '\n'
            '5'
        )

        with self._check_assertion(msg):
            self.assertSpyReturned(obj.do_math.calls[0], 100)

    def test_assertSpyLastReturned_with_expected_return(self):
        """Testing SpyAgency.assertSpyLastReturned with expected return value
        """
        obj = MathClass()
        self.spy_on(obj.do_math)

        obj.do_math(1, b=4)
        obj.do_math(2, b=9)

        # These should not fail.
        self.assertSpyLastReturned(obj.do_math, 11)
        self.assertSpyLastReturned(obj.do_math.spy, 11)

        # Check the aliases.
        self.assert_spy_last_returned(obj.do_math, 11)
        kgb.asserts.assert_spy_last_returned(obj.do_math, 11)

    def test_assertSpyLastReturned_without_expected_return(self):
        """Testing SpyAgency.assertSpyLastReturned without expected return
        value
        """
        obj = MathClass()
        self.spy_on(obj.do_math)

        obj.do_math(1, b=4)
        obj.do_math(2, b=9)

        msg = (
            'The last call to do_math did not return 5.\n'
            '\n'
            'It last returned:\n'
            '\n'
            '11'
        )

        with self._check_assertion(msg):
            self.assertSpyLastReturned(obj.do_math, 5)

        with self._check_assertion(msg):
            self.assertSpyLastReturned(obj.do_math.spy, 5)

    def test_assertSpyRaised_with_expected_exception(self):
        """Testing SpyAgency.assertSpyRaised with expected exception raised"""
        def _do_math(_self, a, *args, **kwargs):
            if a == 1:
                raise KeyError
            elif a == 2:
                raise ValueError

        obj = MathClass()
        self.spy_on(obj.do_math, call_fake=_do_math)

        try:
            obj.do_math(1)
        except KeyError:
            pass

        try:
            obj.do_math(2)
        except ValueError:
            pass

        # These should not fail.
        self.assertSpyRaised(obj.do_math, KeyError)
        self.assertSpyRaised(obj.do_math.calls[0], KeyError)
        self.assertSpyRaised(obj.do_math.spy, ValueError)
        self.assertSpyRaised(obj.do_math.spy.calls[1], ValueError)

        # Check the aliases.
        self.assert_spy_raised(obj.do_math, KeyError)
        kgb.asserts.assert_spy_raised(obj.do_math, KeyError)

    def test_assertSpyRaised_with_expected_no_exception(self):
        """Testing SpyAgency.assertSpyRaised with expected completions without
        exceptions
        """
        obj = MathClass()
        self.spy_on(obj.do_math)

        obj.do_math(1)
        obj.do_math(2)

        # These should not fail.
        self.assertSpyRaised(obj.do_math, None)
        self.assertSpyRaised(obj.do_math.calls[0], None)
        self.assertSpyRaised(obj.do_math.spy, None)
        self.assertSpyRaised(obj.do_math.spy.calls[1], None)

    def test_assertSpyRaised_without_expected_exception(self):
        """Testing SpyAgency.assertSpyRaised without expected exception raised
        """
        def _do_math(_self, a, *args, **kwargs):
            if a == 1:
                raise KeyError
            elif a == 2:
                raise ValueError

        obj = MathClass()
        self.spy_on(obj.do_math, call_fake=_do_math)

        # First test without any exceptions raised
        try:
            obj.do_math(1)
        except KeyError:
            pass

        try:
            obj.do_math(2)
        except ValueError:
            pass

        msg = (
            'No call to do_math raised AttributeError.\n'
            '\n'
            'The following exceptions have been raised:\n'
            '\n'
            'Call 0:\n'
            '  KeyError\n'
            '\n'
            'Call 1:\n'
            '  ValueError'
        )

        with self._check_assertion(msg):
            self.assertSpyRaised(obj.do_math, AttributeError)

        with self._check_assertion(msg):
            self.assertSpyRaised(obj.do_math.spy, AttributeError)

        msg = (
            'This call to do_math did not raise AttributeError. It raised '
            'KeyError.'
        )

        with self._check_assertion(msg):
            self.assertSpyRaised(obj.do_math.calls[0], AttributeError)

    def test_assertSpyRaised_without_raised(self):
        """Testing SpyAgency.assertSpyRaised without any exceptions raised"""
        obj = MathClass()
        self.spy_on(obj.do_math)

        # First test without any exceptions raised
        obj.do_math(1)
        obj.do_math(2)

        msg = 'No call to do_math raised an exception.'

        with self._check_assertion(msg):
            self.assertSpyRaised(obj.do_math, AttributeError)

        with self._check_assertion(msg):
            self.assertSpyRaised(obj.do_math.spy, AttributeError)

        msg = 'This call to do_math did not raise an exception.'

        with self._check_assertion(msg):
            self.assertSpyRaised(obj.do_math.spy.calls[0], AttributeError)

    def test_assertSpyLastRaised_with_expected_exception(self):
        """Testing SpyAgency.assertSpyLastRaised with expected exception
        raised
        """
        def _do_math(_self, a, *args, **kwargs):
            if a == 1:
                raise KeyError
            elif a == 2:
                raise ValueError

        obj = MathClass()
        self.spy_on(obj.do_math, call_fake=_do_math)

        try:
            obj.do_math(1)
        except KeyError:
            pass

        try:
            obj.do_math(2)
        except ValueError:
            pass

        # These should not fail.
        self.assertSpyLastRaised(obj.do_math, ValueError)
        self.assertSpyLastRaised(obj.do_math.spy, ValueError)

        # Check the aliases.
        self.assert_spy_last_raised(obj.do_math, ValueError)
        kgb.asserts.assert_spy_last_raised(obj.do_math, ValueError)

    def test_assertSpyLastRaised_with_expected_no_exception(self):
        """Testing SpyAgency.assertSpyLastRaised with expected completion
        without raising
        """
        obj = MathClass()
        self.spy_on(obj.do_math)

        obj.do_math(1)
        obj.do_math(2)

        # These should not fail.
        self.assertSpyLastRaised(obj.do_math, None)
        self.assertSpyLastRaised(obj.do_math.spy, None)

    def test_assertSpyLastRaised_without_expected_exception(self):
        """Testing SpyAgency.assertSpyLastRaised without expected exception
        raised
        """
        def _do_math(_self, a, *args, **kwargs):
            if a == 1:
                raise KeyError
            elif a == 2:
                raise ValueError

        obj = MathClass()
        self.spy_on(obj.do_math, call_fake=_do_math)

        try:
            obj.do_math(1)
        except KeyError:
            pass

        try:
            obj.do_math(2)
        except ValueError:
            pass

        msg = (
            'The last call to do_math did not raise KeyError. It last '
            'raised ValueError.'
        )

        with self._check_assertion(msg):
            self.assertSpyLastRaised(obj.do_math, KeyError)

        with self._check_assertion(msg):
            self.assertSpyLastRaised(obj.do_math.spy, KeyError)

    def test_assertSpyLastRaised_without_raised(self):
        """Testing SpyAgency.assertSpyLastRaised without exception raised"""
        obj = MathClass()
        self.spy_on(obj.do_math)

        obj.do_math(1)
        obj.do_math(2)

        msg = 'The last call to do_math did not raise an exception.'

        with self._check_assertion(msg):
            self.assertSpyLastRaised(obj.do_math, KeyError)

        with self._check_assertion(msg):
            self.assertSpyLastRaised(obj.do_math.spy, KeyError)

    def test_assertSpyRaisedMessage_with_expected(self):
        """Testing SpyAgency.assertSpyRaised with expected exception and
        message raised
        """
        def _do_math(_self, a, *args, **kwargs):
            if a == 1:
                raise AttributeError('Bad key!')
            elif a == 2:
                raise ValueError('Bad value!')

        obj = MathClass()
        self.spy_on(obj.do_math, call_fake=_do_math)

        try:
            obj.do_math(1)
        except AttributeError:
            pass

        try:
            obj.do_math(2)
        except ValueError:
            pass

        # These should not fail.
        self.assertSpyRaisedMessage(obj.do_math, AttributeError, 'Bad key!')
        self.assertSpyRaisedMessage(obj.do_math.calls[0], AttributeError,
                                    'Bad key!')
        self.assertSpyRaisedMessage(obj.do_math.spy, ValueError, 'Bad value!')
        self.assertSpyRaisedMessage(obj.do_math.spy.calls[1], ValueError,
                                    'Bad value!')

        # Check the aliases.
        self.assert_spy_raised_message(obj.do_math, AttributeError, 'Bad key!')
        kgb.asserts.assert_spy_raised_message(obj.do_math, AttributeError,
                                              'Bad key!')

    def test_assertSpyRaisedMessage_without_expected(self):
        """Testing SpyAgency.assertSpyRaisedMessage without expected exception
        and message raised
        """
        def _do_math(_self, a, *args, **kwargs):
            if a == 1:
                raise AttributeError('Bad key!')
            elif a == 2:
                raise ValueError('Bad value!')

        obj = MathClass()
        self.spy_on(obj.do_math, call_fake=_do_math)

        try:
            obj.do_math(1)
        except AttributeError:
            pass

        try:
            obj.do_math(2)
        except ValueError:
            pass

        # Note that we may end up with different string types with different
        # prefixes on different versions of Python, so we need to repr these.
        msg = (
            'No call to do_math raised AttributeError with message %r.\n'
            '\n'
            'The following exceptions have been raised:\n'
            '\n'
            'Call 0:\n'
            '  exception=AttributeError\n'
            '  message=%r\n'
            '\n'
            'Call 1:\n'
            '  exception=ValueError\n'
            '  message=%r'
            % ('Bad key...', str('Bad key!'), str('Bad value!'))
        )

        with self._check_assertion(msg):
            self.assertSpyRaisedMessage(obj.do_math, AttributeError,
                                        'Bad key...')

        with self._check_assertion(msg):
            self.assertSpyRaisedMessage(obj.do_math.spy, AttributeError,
                                        'Bad key...')

        msg = (
            'This call to do_math did not raise AttributeError with message'
            ' %r.\n'
            '\n'
            'It raised:\n'
            '\n'
            'exception=AttributeError\n'
            'message=%r'
            % ('Bad key...', str('Bad key!'))
        )

        with self._check_assertion(msg):
            self.assertSpyRaisedMessage(obj.do_math.calls[0], AttributeError,
                                        'Bad key...')

    def test_assertSpyRaisedMessage_without_raised(self):
        """Testing SpyAgency.assertSpyRaisedMessage without exception raised
        """
        obj = MathClass()
        self.spy_on(obj.do_math)

        obj.do_math(1)
        obj.do_math(2)

        msg = 'No call to do_math raised an exception.'

        with self._check_assertion(msg):
            self.assertSpyRaisedMessage(obj.do_math, KeyError, '...')

        with self._check_assertion(msg):
            self.assertSpyRaisedMessage(obj.do_math.spy, KeyError, '...')

    def test_assertSpyLastRaisedMessage_with_expected(self):
        """Testing SpyAgency.assertSpyLastRaised with expected exception and
        message raised
        """
        def _do_math(_self, a, *args, **kwargs):
            if a == 1:
                raise AttributeError('Bad key!')
            elif a == 2:
                raise ValueError('Bad value!')

        obj = MathClass()
        self.spy_on(obj.do_math, call_fake=_do_math)

        try:
            obj.do_math(1)
        except AttributeError:
            pass

        try:
            obj.do_math(2)
        except ValueError:
            pass

        # These should not fail.
        self.assertSpyLastRaisedMessage(obj.do_math, ValueError, 'Bad value!')
        self.assertSpyLastRaisedMessage(obj.do_math.spy, ValueError,
                                        'Bad value!')

        # Check the aliases.
        self.assert_spy_last_raised_message(obj.do_math, ValueError,
                                            'Bad value!')
        kgb.asserts.assert_spy_last_raised_message(obj.do_math, ValueError,
                                                   'Bad value!')

    def test_assertSpyLastRaisedMessage_without_expected(self):
        """Testing SpyAgency.assertSpyLastRaisedMessage without expected
        exception and message raised
        """
        def _do_math(_self, a, *args, **kwargs):
            if a == 1:
                raise AttributeError('Bad key!')
            elif a == 2:
                raise ValueError('Bad value!')

        obj = MathClass()
        self.spy_on(obj.do_math, call_fake=_do_math)

        try:
            obj.do_math(1)
        except AttributeError:
            pass

        try:
            obj.do_math(2)
        except ValueError:
            pass

        # Note that we may end up with different string types with different
        # prefixes on different versions of Python, so we need to repr these.
        msg = (
            'The last call to do_math did not raise AttributeError with '
            'message %r.\n'
            '\n'
            'It last raised:\n'
            '\n'
            'exception=ValueError\n'
            'message=%r'
            % ('Bad key!', str('Bad value!'))
        )

        with self._check_assertion(msg):
            self.assertSpyLastRaisedMessage(obj.do_math, AttributeError,
                                            'Bad key!')

        with self._check_assertion(msg):
            self.assertSpyLastRaisedMessage(obj.do_math.spy, AttributeError,
                                            'Bad key!')

    def test_assertSpyLastRaisedMessage_without_raised(self):
        """Testing SpyAgency.assertSpyLastRaisedMessage without exception
        raised
        """
        obj = MathClass()
        self.spy_on(obj.do_math)

        obj.do_math(1)
        obj.do_math(2)

        msg = 'The last call to do_math did not raise an exception.'

        with self._check_assertion(msg):
            self.assertSpyLastRaisedMessage(obj.do_math, KeyError, '...')

        with self._check_assertion(msg):
            self.assertSpyLastRaisedMessage(obj.do_math.spy, KeyError, '...')

    @contextmanager
    def _check_assertion(self, msg):
        """Check that the expected assertion and message is raised.

        Args:
            msg (unicode):
                The assertion message.

        Context:
            The context used to run an assertion.
        """
        with self.assertRaises(AssertionError) as ctx:
            yield

        self.assertEqual(str(ctx.exception), msg)



"""Common utility functions used by kgb."""

from __future__ import unicode_literals

import inspect
from unittest.util import safe_repr

from kgb.pycompat import iteritems


def get_defined_attr_value(owner, name, ancestors_only=False):
    """Return a value as defined in a class, instance, or ancestor.

    This will look for the real definition, and not the definition returned
    when accessing the attribute or instantiating a class. This will bypass any
    descriptors and return the actual definition from the instance or class
    that defines it.

    Args:
        owner (type or object):
            The owner of the attribute.

        name (unicode):
            The name of the attribute.

        ancestors_only (bool, optional):
            Whether to only look in ancestors of ``owner``, and not in
            ``owner`` itself.

    Returns:
        object:
        The attribute value.

    Raises:
        AttributeError:
            The attribute could not be found.
    """
    if not ancestors_only:
        d = owner.__dict__

        if name in d:
            return d[name]

    if not inspect.isclass(owner):
        # NOTE: It's important we use __class__ and not type(). They are not
        #       synonymous. The latter will not return the class for an
        #       instance for old-style classes.
        return get_defined_attr_value(owner.__class__, name)

    for parent_cls in owner.__bases__:
        try:
            return get_defined_attr_value(parent_cls, name)
        except AttributeError:
            pass

    raise AttributeError


def is_attr_defined_on_ancestor(cls, name):
    """Return whether an attribute is defined on an ancestor of a class.

    Args:
        name (unicode):
            The name of the attribute.

    Returns:
        bool:
        ``True`` if an ancestor defined the attribute. ``False`` if it did not.
    """
    try:
        get_defined_attr_value(cls, name, ancestors_only=True)
        return True
    except AttributeError:
        return False


def format_spy_kwargs(kwargs):
    """Format keyword arguments.

    This will convert all keys to native strings, to help with showing
    more reasonable output that's consistent. The keys will also be
    provided in sorted order.

    Args:
        kwargs (dict):
            The dictionary of keyword arguments.

    Returns:
        unicode:
        The formatted string representation.
    """
    return '{%s}' % ', '.join(
        '%s: %s' % (safe_repr(str(key)), safe_repr(value))
        for key, value in sorted(iteritems(kwargs),
                                 key=lambda pair: pair[0])
    )



from __future__ import absolute_import, unicode_literals

import copy
import inspect
import types

from kgb.calls import SpyCall
from kgb.errors import (ExistingSpyError,
                        IncompatibleFunctionError,
                        InternalKGBError)
from kgb.pycompat import iterkeys, pyver
from kgb.signature import FunctionSig, _UNSET_ARG
from kgb.utils import is_attr_defined_on_ancestor


class FunctionSpy(object):
    """A spy infiltrating a function.

    A FunctionSpy takes the place of another function. It will record any
    calls made to the function for later inspection.

    By default, a FunctionSpy will allow the call to go through to the
    original function. This can be disabled by passing call_original=False
    when initiating the spy. If disabled, the original function will never be
    called.

    This can also be passed a call_fake parameter pointing to another
    function to call instead of the original. If passed, this will take
    precedence over call_original.
    """

    #: The spy represents a standard function.
    TYPE_FUNCTION = FunctionSig.TYPE_FUNCTION

    #: The spy represents a bound method.
    #:
    #: Bound methods are functions on an instance of a class, or classmethods.
    TYPE_BOUND_METHOD = FunctionSig.TYPE_BOUND_METHOD

    #: The spy represents an unbound method.
    #:
    #: Unbound methods are standard methods on a class.
    TYPE_UNBOUND_METHOD = FunctionSig.TYPE_UNBOUND_METHOD

    _PROXY_METHODS = [
        'call_original', 'called_with', 'last_called_with',
        'raised', 'last_raised', 'returned', 'last_returned',
        'raised_with_message', 'last_raised_with_message',
        'reset_calls', 'unspy',
    ]

    _FUNC_ATTR_DEFAULTS = {
        'calls': [],
        'called': False,
        'last_call': None,
    }

    _spy_map = {}

    def __init__(self, agency, func, call_fake=None, call_original=True,
                 op=None, owner=_UNSET_ARG, func_name=None):
        """Initialize the spy.

        This will begin spying on the provided function or method, injecting
        new code into the function to help record how it was called and
        what it returned, and adding methods and state onto the function
        for callers to access in order to get those results.

        Version Added:
            7.0:
            Added support for specifying an explicit function name using
            ``func_name=``.

        Version Added:
            5.0:
            Added support for specifying an instance in ``owner`` when spying
            on bound methods using decorators that return plain functions.

        Args:
            agency (kgb.agency.SpyAgency):
                The spy agency that manages this spy.

            func (callable):
                The function or method to spy on.

            call_fake (callable, optional):
                The optional function to call when this function is invoked.

                This cannot be specified if ``op`` is provided.

            call_original (bool, optional):
                Whether to call the original function when the spy is
                invoked. If ``False``, no function will be called.

                This is ignored if ``call_fake`` or ``op`` are provided.

            op (kgb.spies.BaseOperation, optional):
                An operation to perform.

                This cannot be specified if ``call_fake`` is provided.

            owner (type or object, optional):
                The owner of the function or method.

                If spying on an unbound method, this **must** be set to the
                class that owns it.

                If spying on a bound method that identifies as a plain
                function (which may happen if the method is decorated and
                dynamically returns a new function on access), this should
                be the instance of the object you're spying on.

            func_name (str, optional):
                An explicit name for the function. This will be used instead
                of the function's specified name, and is usually a sign of a
                bad decorator.

                Version Added:
                    7.0
        """
        # Start off by grabbing the current frame. This will be needed for
        # some errors.
        self.init_frame = inspect.currentframe()

        # Check the parameters passed to make sure that invalid data wasn't
        # provided.
        if op is not None and call_fake is not None:
            raise ValueError('op and call_fake cannot both be provided.')

        if hasattr(func, 'spy'):
            raise ExistingSpyError(func)

        if (not callable(func) or
            not hasattr(func, FunctionSig.FUNC_NAME_ATTR) or
            not (hasattr(func, FunctionSig.METHOD_SELF_ATTR) or
                 hasattr(func, FunctionSig.FUNC_GLOBALS_ATTR))):
            raise ValueError('%r cannot be spied on. It does not appear to '
                             'be a valid function or method.'
                             % func)

        # Construct a signature for the function and begin closely inspecting
        # the parameters, making sure everything will be compatible so we
        # don't have unexpected breakages when setting up or calling spies.
        sig = FunctionSig(func=func,
                          owner=owner,
                          func_name=func_name)
        self._sig = sig

        # If the caller passed an explicit owner, check to see if it's at all
        # valid. Note that it may have been handled above (for unbound
        # methods).
        if owner is not _UNSET_ARG and owner is not self.owner:
            if self.func_type == self.TYPE_FUNCTION:
                raise ValueError(
                    'This function has no owner, but an owner was passed '
                    'to spy_on().')
            else:
                if not hasattr(owner, self.func_name):
                    raise ValueError('The owner passed does not contain the '
                                     'spied method.')
                elif (self.func_type == self.TYPE_BOUND_METHOD or
                      (pyver[0] == 2 and
                       self.func_type == self.TYPE_UNBOUND_METHOD)):
                    raise ValueError(
                        'The owner passed does not match the actual owner of '
                        'the bound method.')

        # We cannot currently spy on unbound methods that result in slippery
        # functions, so check for that and bail early.
        if (sig.is_slippery and
            self.func_type == self.TYPE_UNBOUND_METHOD):
            raise ValueError('Unable to spy on unbound slippery methods '
                             '(those that return a new function on each '
                             'attribute access). Please spy on an instance '
                             'instead.')

        # If call_fake was provided, check that it's valid and has a
        # compatible function signature.
        if op is not None:
            # We've already checked this above, but check it again.
            assert call_fake is None

            call_fake = op.setup(self)
            assert call_fake is not None

        if call_fake is not None:
            if not callable(call_fake):
                raise ValueError('%r cannot be used for call_fake. It does '
                                 'not appear to be a valid function or method.'
                                 % call_fake)

            call_fake_sig = FunctionSig(call_fake,
                                        func_name=func_name)

            if not sig.is_compatible_with(call_fake_sig):
                raise IncompatibleFunctionError(
                    func=func,
                    func_sig=sig,
                    incompatible_func=call_fake,
                    incompatible_func_sig=call_fake_sig)

        # Now that we're done validating, we can start setting state and
        # patching things.
        self.agency = agency
        self.orig_func = func
        self._real_func = sig.real_func
        self._call_orig_func = self._clone_function(self.orig_func)

        if self._get_owner_needs_patching():
            # We need to store the original attribute value for the function,
            # as defined in the class that owns it. That may be the provided
            # or calculated owner, or a parent of it.
            #
            # This is needed because the function provided may not actually be
            # what's defined on the class. What's defined might be a decorator
            # that returns a function, and it might not even be the same
            # function each time it's accessed.
            self._owner_func_attr_value = \
                self.owner.__dict__.get(self.func_name)

            # Now we can patch the owner to prevent conflicts between spies.
            self._patch_owner()
        else:
            self._owner_func_attr_value = self.orig_func

        # Determine what we're going to invoke when the spy is called.
        if call_fake:
            self.func = call_fake
        elif call_original:
            self.func = self.orig_func
        else:
            self.func = None

        # Build our proxy function. This is the spy itself, the function that
        # will actually be invoked when the spied-on function is called.
        self._build_proxy_func(func)

        # If we're calling the original function above, we need to replace what
        # we're calling with something that acts like the original function.
        # Otherwise, we'll just call the forwarding_call above in an infinite
        # loop.
        if self.func is self.orig_func:
            self.func = self._clone_function(self.func,
                                             code=self._old_code)

    @property
    def func_type(self):
        """The type of function being spied on.

        This will be one of :py:attr:`TYPE_FUNCTION`,
        :py:attr:`TYPE_UNBOUND_METHOD`, or :py:attr:`TYPE_BOUND_METHOD`.

        Type:
            int
        """
        return self._sig.func_type

    @property
    def func_name(self):
        """The name of the function being spied on.

        Type:
            str
        """
        return self._sig.func_name

    @property
    def owner(self):
        """The owner of the method, if a bound or unbound method.

        This will be ``None`` if there is no owner.

        Type:
            type
        """
        return self._sig.owner

    @property
    def called(self):
        """Whether or not the spy was ever called."""
        try:
            return self._real_func.called
        except AttributeError:
            return False

    @property
    def calls(self):
        """The list of calls made to the function.

        Each is an instance of :py:class:`SpyCall`.
        """
        try:
            return self._real_func.calls
        except AttributeError:
            return []

    @property
    def last_call(self):
        """The last call made to this function.

        If a spy hasn't been called yet, this will be ``None``.
        """
        try:
            return self._real_func.last_call
        except AttributeError:
            return None

    def unspy(self, unregister=True):
        """Remove the spy from the function, restoring the original.

        The spy will, by default, be removed from the registry's
        list of spies. This can be disabled by passing ``unregister=False``,
        but don't do that. That's for internal use.

        Args:
            unregister (bool, optional):
                Whether to unregister the spy from the associated agency.
        """
        real_func = self._real_func
        owner = self.owner

        assert hasattr(real_func, 'spy')

        del FunctionSpy._spy_map[id(self)]
        del real_func.spy

        for attr_name in iterkeys(self._FUNC_ATTR_DEFAULTS):
            delattr(real_func, attr_name)

        for func_name in self._PROXY_METHODS:
            delattr(real_func, func_name)

        setattr(real_func, FunctionSig.FUNC_CODE_ATTR, self._old_code)

        if owner is not None:
            self._set_method(owner, self.func_name,
                             self._owner_func_attr_value)

        if unregister:
            self.agency.spies.remove(self)

    def call_original(self, *args, **kwargs):
        """Call the original function being spied on.

        The function will behave as normal, and will not trigger any spied
        behavior or call tracking.

        Args:
            *args (tuple):
                The positional arguments to pass to the function.

            **kwargs (dict):
                The keyword arguments to pass to the function.

        Returns:
            object:
            The return value of the function.

        Raises:
            Exception:
                Any exceptions raised by the function.
        """
        if self.func_type == self.TYPE_BOUND_METHOD:
            return self._call_orig_func(self.owner, *args, **kwargs)
        else:
            if self.func_type == self.TYPE_UNBOUND_METHOD:
                if not args or not isinstance(args[0], self.owner):
                    raise TypeError(
                        'The first argument to %s.call_original() must be '
                        'an instance of %s.%s, since this is an unbound '
                        'method.'
                        % (self._call_orig_func.__name__,
                           self.owner.__module__,
                           self.owner.__name__))

            return self._call_orig_func(*args, **kwargs)

    def called_with(self, *args, **kwargs):
        """Return whether the spy was ever called with the given arguments.

        This will check each and every recorded call to see if the arguments
        and keyword arguments match up. If at least one call does match, this
        will return ``True``.

        Not every argument and keyword argument made in the call must be
        provided to this method. These can be a subset of the positional and
        keyword arguments in the call, but cannot contain any arguments not
        made in the call.

        Args:
            *args (tuple):
                The positional arguments made in the call, or a subset of
                those arguments (starting with the first argument).

            **kwargs (dict):
                The keyword arguments made in the call, or a subset of those
                arguments.

        Returns:
            bool:
            ``True`` if there's at least one call matching these arguments.
            ``False`` if no call matches.
        """
        return any(
            call.called_with(*args, **kwargs)
            for call in self.calls
        )

    def last_called_with(self, *args, **kwargs):
        """Return whether the spy was last called with the given arguments.

        Not every argument and keyword argument made in the call must be
        provided to this method. These can be a subset of the positional and
        keyword arguments in the call, but cannot contain any arguments not
        made in the call.

        Args:
            *args (tuple):
                The positional arguments made in the call, or a subset of
                those arguments (starting with the first argument).

            **kwargs (dict):
                The keyword arguments made in the call, or a subset of those
                arguments.

        Returns:
            bool:
            ``True`` if the last call's arguments match the provided arguments.
            ``False`` if they do not.
        """
        call = self.last_call

        return call is not None and call.called_with(*args, **kwargs)

    def returned(self, value):
        """Return whether the spy was ever called and returned the given value.

        This will check each and every recorded call to see if any of them
        returned the given value.  If at least one call did, this will return
        ``True``.

        Args:
            value (object):
                The expected returned value from the call.

        Returns:
            bool:
            ``True`` if there's at least one call that returned this value.
            ``False`` if no call returned the value.
        """
        return any(
            call.returned(value)
            for call in self.calls
        )

    def last_returned(self, value):
        """Return whether the spy's last call returned the given value.

        Args:
            value (object):
                The expected returned value from the call.

        Returns:
            bool:
            ``True`` if the last call returned this value. ``False`` if it
            did not.
        """
        call = self.last_call

        return call is not None and call.returned(value)

    def raised(self, exception_cls):
        """Return whether the spy was ever called and raised this exception.

        This will check each and every recorded call to see if any of them
        raised an exception of a given type. If at least one call does match,
        this will return ``True``.

        Args:
            exception_cls (type):
                The expected type of exception raised by a call.

        Returns:
            bool:
            ``True`` if there's at least one call raising the given exception
            type. ``False`` if no call matches.
        """
        return any(
            call.raised(exception_cls)
            for call in self.calls
        )

    def last_raised(self, exception_cls):
        """Return whether the spy's last call raised this exception.

        Args:
            exception_cls (type):
                The expected type of exception raised by a call.

        Returns:
            bool:
            ``True`` if the last call raised the given exception type.
            ``False`` if it did not.
        """
        call = self.last_call

        return call is not None and call.raised(exception_cls)

    def raised_with_message(self, exception_cls, message):
        """Return whether the spy's calls ever raised this exception/message.

        This will check each and every recorded call to see if any of them
        raised an exception of a given type with the given message. If at least
        one call does match, this will return ``True``.

        Args:
            exception_cls (type):
                The expected type of exception raised by a call.

            message (unicode):
                The expected message from the exception.

        Returns:
            bool:
            ``True`` if there's at least one call raising the given exception
            type and message. ``False`` if no call matches.
        """
        return any(
            call.raised_with_message(exception_cls, message)
            for call in self.calls
        )

    def last_raised_with_message(self, exception_cls, message):
        """Return whether the spy's last call raised this exception/message.

        Args:
            exception_cls (type):
                The expected type of exception raised by a call.

            message (unicode):
                The expected message from the exception.

        Returns:
            bool:
            ``True`` if the last call raised the given exception type and
            message. ``False`` if it did not.
        """
        call = self.last_call

        return (call is not None and
                call.raised_with_message(exception_cls, message))

    def reset_calls(self):
        """Reset the list of calls recorded by this spy."""
        self._real_func.calls = []
        self._real_func.called = False
        self._real_func.last_call = None

    def __call__(self, *args, **kwargs):
        """Call the original function or fake function for the spy.

        This will be called automatically when calling the spied function,
        recording the call and the results from the call.

        Args:
            *args (tuple):
                Positional arguments passed to the function.

            **kwargs (dict):
                All dictionary arguments either passed to the function or
                default values for unspecified keyword arguments in the
                function signature.

        Returns:
            object:
            The result of the function call.
        """
        record_args = args

        if self.func_type in (self.TYPE_BOUND_METHOD,
                              self.TYPE_UNBOUND_METHOD):
            record_args = record_args[1:]

        sig = self._sig
        real_func = self._real_func
        func = self.func

        call = SpyCall(self, record_args, kwargs)
        real_func.calls.append(call)
        real_func.called = True
        real_func.last_call = call

        if func is None:
            result = None
        else:
            try:
                if sig.has_getter:
                    # This isn't a standard function. It's a descriptor with
                    # a __get__() method. We need to fetch the value it
                    # returns.
                    result = sig.defined_func.__get__(self.owner)

                    if sig.is_slippery:
                        # Since we know this represents a slippery function,
                        # we need to take the function from the descriptor's
                        # result and call it.
                        result = result(*args, **kwargs)
                else:
                    # This is a typical function/method. We can call it
                    # directly.
                    result = func(*args, **kwargs)
            except Exception as e:
                call.exception = e
                raise

            call.return_value = result

        return result

    def __repr__(self):
        """Return a string representation of the spy.

        This is mainly used for debugging information. It will show some
        details on the spied function and call log.

        Returns:
            unicode:
            The resulting string representation.
        """
        func_type = self.func_type

        if func_type == self.TYPE_FUNCTION:
            func_type_str = 'function'
            qualname = self.func_name
        else:
            owner = self.owner

            if func_type == self.TYPE_BOUND_METHOD:
                # It's important we use __class__ instead of type(), because
                # we may be dealing with an old-style class.
                owner_cls = self.owner.__class__

                if owner_cls is type:
                    class_name = owner.__name__
                    func_type_str = 'classmethod'
                else:
                    class_name = owner_cls.__name__
                    func_type_str = 'bound method'
            elif func_type == self.TYPE_UNBOUND_METHOD:
                class_name = owner.__name__
                func_type_str = 'unbound method'

            qualname = '%s.%s of %r' % (class_name, self.func_name, owner)

        call_count = len(self.calls)

        if call_count == 1:
            calls_str = 'call'
        else:
            calls_str = 'calls'

        return '<Spy for %s %s (%d %s)>' % (func_type_str, qualname,
                                            len(self.calls), calls_str)

    def _get_owner_needs_patching(self):
        """Return whether the owner (if any) needs to be patched.

        Owners need patching if they're an instance, if the function is
        slippery, or if the function is defined on an ancestor of the class
        and not the class itself.

        See :py:meth:`_patch_owner` for what patching entails.

        Returns:
            bool:
            ``True`` if the owner needs patching. ``False`` if it does not.
        """
        owner = self.owner

        return (owner is not None and
                (not inspect.isclass(owner) or
                 self._sig.is_slippery or
                 is_attr_defined_on_ancestor(owner, self.func_name)))

    def _patch_owner(self):
        """Patch the owner.

        This will create a new method in place of an existing one on the
        owner, in order to ensure that the owner has its own unique copy
        for spying purposes.

        Patching the owner will avoid collisions between spies in the event
        that the method being spied on is defined by a parent of the owner,
        rather than the owner itself.

        See :py:meth:`_get_owner_needs_patching` the conditions under which
        patching will occur.
        """
        # Construct a replacement function for this method, and
        # re-assign it to the owner. We do this in order to prevent
        # two spies on the same method on two separate instances
        # of the class, or two subclasses of a common class owning the
        # method from conflicting with each other.
        real_func = self._clone_function(self._real_func)
        owner = self.owner

        if self.func_type == self.TYPE_BOUND_METHOD:
            method_type_args = [real_func, owner]

            if pyver[0] >= 3:
                method_type_args.append(owner)

            self._set_method(owner, self.func_name,
                             types.MethodType(real_func, self.owner))
        else:
            self._set_method(owner, self.func_name, real_func)

        self._real_func = real_func

    def _build_proxy_func(self, func):
        """Build the proxy function used to forward calls to this spy.

        This will construct a new function compatible with the signature of
        the provided function, which will call this spy whenever it's called.
        The bytecode of the provided function will be set to that of the
        generated proxy function. See the comment within this function for
        details on how this works.

        Args:
            func (callable):
                The function to proxy.
        """
        # Prior to kgb 2.0, we attempted to optimistically replace
        # methods on a class with a FunctionSpy, forwarding on calls to the
        # fake or original function. This was the design since kgb 1.0, but
        # wasn't sufficient. We realized in the first release that this
        # wouldn't work for standard functions, and so we had two designs:
        # One for methods, one for standard functions.
        #
        # In kgb 2.0, in an effort to standardize behavior, we moved fully
        # to the method originally used for standard functions (largely due
        # to the fact that in Python 3, unbound methods are just standard
        # functions).
        #
        # Standard functions can't be replaced. Unlike a bound function,
        # we can't reliably figure out what dictionary it lives in (it
        # could be a locals() inside another function), and even if we
        # replace that, we can't replace all the copies that have been
        # imported up to this point.
        #
        # The only option is to change what happens when we call the
        # function. That's easier said than done. We can't just replace
        # the __call__ method on it, like you could on a fake method for
        # a class.
        #
        # What we must do is replace the code backing it. This must be
        # done carefully. The "co_freevars" and "co_cellvars" fields must
        # remain the same between the old code and the new one. The
        # actual bytecode and most of the rest of the fields can be taken
        # from another function (the "forwarding_call" function defined
        # inline below).
        #
        # Unfortunately, we no longer have access to "self" (since we
        # replaced "co_freevars"). Instead, we store a global mapping
        # of codes to spies.
        #
        # We also must build the function dynamically, using exec().
        # The reason is that we want to accurately mimic the function
        # signature of the original function (in terms of specifying
        # the correct positional and keyword arguments). The way we format
        # arguments depends on the version of Python. We maintain
        # compatibility through the FunctionSig.format_arg_spec() methods
        # (which has implementations for both Python 2 and 3).
        #
        # We do use different values for the default keyword arguments,
        # which is actually okay. Within the function, these will all be
        # set to a special value (_UNSET_ARG), which is used later for
        # determining which keyword arguments were provided and which
        # were not. Anything attempting to inspect this function with
        # getargspec(), getfullargspec(), or inspect.Signature will get the
        # defaults from the original function, by way of the
        # original func.func_defaults attribute (on Python 2) or
        # __defaults__ (on Python 3).
        #
        # This forwarding function then needs to call the forwarded
        # function in exactly the same manner as it was called. That is,
        # if a keyword argument's value was passed in as a positional
        # argument, or a positional argument was specified as a keyword
        # argument in the call, then the forwarded function must be
        # called the same way, for argument tracking and signature
        # compatibility.
        #
        # In order to do this, we have to find out how forwarding_call was
        # called. This can be done by inspecting the bytecode of the
        # call in the parent frame and getting the number of positional
        # and keyword arguments used. From there, we can determine which
        # argument slots were specified and start looking for any keyword
        # arguments not set to _UNSET_ARG, passing them through to the
        # original function in the same order. Doing this requires
        # another exec() call in order to build out those arguments.
        #
        # Within the function, all imports and variables are prefixed to
        # avoid the possibility of collisions with arguments.
        #
        # Since we're only overriding the code, all other attributes (like
        # func_defaults, __doc__, etc.) will make use of those from
        # the original function.
        #
        # The result is that we've completely hijacked the original
        # function, making it call our own forwarding function instead.
        # It's a wonderful bag of tricks that are fully legal, but really
        # dirty. Somehow, it all really fits in with the idea of spies,
        # though.
        sig = self._sig
        spy_id = id(self)
        real_func = self._real_func

        forwarding_call = self._compile_forwarding_call_func(
            func=func,
            sig=sig,
            spy_id=spy_id)

        old_code, new_code = self._build_spy_code(func, forwarding_call)
        self._old_code = old_code
        setattr(real_func, FunctionSig.FUNC_CODE_ATTR, new_code)

        # Update our spy lookup map so the proxy function can easily find
        # the spy instance.
        FunctionSpy._spy_map[spy_id] = self

        # Update the attributes on the function. we'll be placing all spy
        # state and some proxy methods pointing to this spy, so that we can
        # easily access them through the function.
        real_func.spy = self
        real_func.__dict__.update(copy.deepcopy(self._FUNC_ATTR_DEFAULTS))

        for proxy_func_name in self._PROXY_METHODS:
            assert not hasattr(real_func, proxy_func_name)
            setattr(real_func, proxy_func_name, getattr(self, proxy_func_name))

    def _compile_forwarding_call_func(self, func, sig, spy_id):
        """Compile a forwarding call function for the spy.

        This will build the Python code for a function that approximates the
        function we're spying on, with the same function definition and
        closure behavior.

        Version Added:
            7.1

        Args:
            func (callable):
                The function being spied on.

            sig (kgb.signature.BaseFunctionSig):
                The function signature to use for this function.

            spy_id (int):
                The ID used for the spy registration.

        Returns:
            callable:
            The resulting forwarding function.
        """
        closure_vars = func.__code__.co_freevars
        use_closure = bool(closure_vars)

        # If the function is in a closure, we'll need to mirror the closure
        # state by using the referenced variables within _kgb_forwarding_call
        # and by defining those variables within a closure.
        #
        # Start by setting up a string that will use each closure.
        if use_closure:
            # This is an efficient way of referencing each variable without
            # side effects (at least in Python 2.7 through 3.11). Tuple
            # operations are fast and compact, and don't risk any inadvertent
            # invocation of the variables.
            use_closure_vars_str = (
                '        (%s)\n'
                % ', '.join(func.__code__.co_freevars)
            )
        else:
            # No closure, so nothing to set up.
            use_closure_vars_str = ''

        # Now define the forwarding call. This will always be nested within
        # either a closure of an if statement, letting us build a single
        # version at the right indentation level, keeping this as fast and
        # portable as possible.
        forwarding_call_str = (
            '    def _kgb_forwarding_call(%(params)s):\n'
            '        from kgb.spies import FunctionSpy as _kgb_cls\n'
            '%(use_closure_vars)s'
            '        _kgb_l = locals()\n'
            '        return _kgb_cls._spy_map[%(spy_id)s](%(call_args)s)\n'
            % {
                'call_args': sig.format_forward_call_args(),
                'params': sig.format_arg_spec(),
                'spy_id': spy_id,
                'use_closure_vars': use_closure_vars_str,
            }
        )

        if use_closure:
            # We now need to put _kgb_forwarding_call in a closure, to mirror
            # the behavior of the spied function. The closure will provide
            # the closure variables, and will return the function we can
            # later use.
            func_code_str = (
                'def _kgb_forwarding_call_closure(%(params)s):\n'
                '%(forwarding_call)s'
                '    return _kgb_forwarding_call\n'
                % {
                    'forwarding_call': forwarding_call_str,
                    'params': ', '.join(
                        '%s=None' % _var
                        for _var in closure_vars
                    )
                }
            )
        else:
            # No closure, so just define the function as-is. We will need to
            # wrap in an "if 1:" though, just to ensure indentation is fine.
            func_code_str = (
                'if 1:\n'
                '%s'
                % forwarding_call_str
            )

        # We can now build our function.
        exec_locals = {}

        try:
            eval(compile(func_code_str, '<string>', 'exec'),
                 globals(), exec_locals)
        except Exception as e:
            raise InternalKGBError(
                'Unable to compile a spy function for %(func)r: %(error)s'
                '\n\n'
                '%(code)s'
                % {
                    'code': func_code_str,
                    'error': e,
                    'func': func,
                })

        # Grab the resulting compiled function out of the locals.
        if use_closure:
            # It's in our closure, so call that and get the result.
            forwarding_call = exec_locals['_kgb_forwarding_call_closure']()
        else:
            forwarding_call = exec_locals['_kgb_forwarding_call']

        assert forwarding_call is not None

        return forwarding_call

    def _build_spy_code(self, func, forwarding_call):
        """Build a CodeType to inject into the spied function.

        This will create a function bytecode object that contains a mix of
        attributes from the original function and the forwarding call. The
        result can be injected directly into the spied function, containing
        just the right data to impersonate the function and call our own
        logic.

        Version Added:
            7.1

        Args:
            func (callable):
                The function being spied on.

            forwarding_call (callable):
                The spy forwarding call we built.

        Returns:
            tuple:
            A 2-tuple containing:

            1. The spied function's code object (:py:class:`types.CodeType`).
            1. The new spy code object (:py:class:`types.CodeType`).
        """
        old_code = getattr(func, FunctionSig.FUNC_CODE_ATTR)
        temp_code = getattr(forwarding_call, FunctionSig.FUNC_CODE_ATTR)

        assert old_code != temp_code

        if hasattr(old_code, 'replace'):
            # Python >= 3.8
            #
            # It's important we replace the code instead of building a new
            # one when possible. On Python 3.11, this will ensure that
            # state needed for exceptions (co_positions()) will be set
            # correctly.
            #
            # NOTE: Prior to kgb 7.2, we had set co_freevars and co_vellvars
            #       here. This caused crashes with Python 3.13 beta 2 (the
            #       latest release as of this writing -- June 19, 2024). We
            #       don't appear to actually need or want to set these on
            #       Python 3, so we removed this.
            replace_kwargs = {
                'co_name': old_code.co_name,
            }

            if pyver >= (3, 11):
                replace_kwargs['co_qualname'] = old_code.co_qualname

            # Python 3.13+ includes a warning when assigning a code object if
            # the generator flag doesn't match. The actual bug that caused them
            # to add this warning can be found here:
            #
            #     https://github.com/python/cpython/issues/81137
            #
            # This is a pretty isolated, unique case where the user was
            # replacing the code object for a lambda with the gi_code from a
            # generator expression, causing CPython to crash. This is purely
            # an issue with the gi_code object, and does not affect generator
            # functions (or even a lambda that returns a generator expression)
            # in the same way, so kgb's use is fine. We therefore mirror these
            # flags onto our replacement code object to prevent the warning.
            #
            # Our forwarding function won't have any of these flags set, so we
            # can just copy the positive case for all of them if they exist on
            # the function we're replacing.
            #
            # We only do this on 3.11+ because on earlier versions this causes
            # the spy to not execute.
            if pyver >= (3, 11):
                mirror_flags = old_code.co_flags & (
                    inspect.CO_GENERATOR |
                    inspect.CO_COROUTINE |
                    inspect.CO_ASYNC_GENERATOR)
                replace_kwargs['co_flags'] = temp_code.co_flags | mirror_flags

            new_code = temp_code.replace(**replace_kwargs)
        else:
            # Python <= 3.7
            #
            # We have to build this manually, using a combination of the
            # two. We won't bother with anything newer than Python 3.7.
            code_args = [temp_code.co_argcount]

            if pyver >= (3, 0):
                code_args.append(temp_code.co_kwonlyargcount)

            code_args += [
                temp_code.co_nlocals,
                temp_code.co_stacksize,
                temp_code.co_flags,
                temp_code.co_code,
                temp_code.co_consts,
                temp_code.co_names,
                temp_code.co_varnames,
                temp_code.co_filename,
                old_code.co_name,
                temp_code.co_firstlineno,
                temp_code.co_lnotab,
                old_code.co_freevars,
                old_code.co_cellvars,
            ]

            new_code = types.CodeType(*code_args)

        assert new_code != old_code
        assert new_code != temp_code

        return old_code, new_code

    def _clone_function(self, func, code=None):
        """Clone a function, optionally providing new bytecode.

        This will create a new function that contains all the state of the
        original (including annotations and any default argument values).

        Args:
            func (types.FunctionType):
                The function to clone.

            code (types.CodeType, optional):
                The new bytecode for the function. If not specified, the
                original function's bytecode will be used.

        Returns:
            types.FunctionType:
            The new function.
        """
        cloned_func = types.FunctionType(
            code or getattr(func, FunctionSig.FUNC_CODE_ATTR),
            getattr(func, FunctionSig.FUNC_GLOBALS_ATTR),
            getattr(func, FunctionSig.FUNC_NAME_ATTR),
            getattr(func, FunctionSig.FUNC_DEFAULTS_ATTR),
            getattr(func, FunctionSig.FUNC_CLOSURE_ATTR))

        if pyver[0] >= 3:
            # Python 3.x doesn't support providing any of the new
            # metadata introduced in Python 3.x to the constructor of
            # FunctionType. We have to set those manually.
            for attr in ('__annotations__', '__kwdefaults__'):
                setattr(cloned_func, attr, copy.deepcopy(getattr(func, attr)))

        return cloned_func

    def _set_method(self, owner, name, method):
        """Set a new method on an object.

        This will set the method (or delete the attribute for one if setting
        ``None``).

        If setting on a class, this will use a standard
        :py:func:`setattr`/:py:func:`delattr`.

        If setting on an instance, this will use a standard
        :py:meth:`object.__setattr__`/:py:meth:`object.__delattr__` (in order
        to avoid triggering a subclass-defined version of
        :py:meth:`~object.__setattr__`/:py:meth:`~object.__delattr__`, which
        might lose or override our spy).

        Args:
            owner (type or object):
                The class or instance to set the method on.

            name (unicode):
                The name of the attribute to set for the method.

            method (types.MethodType):
                The method to set (or ``None`` to delete).
        """
        if inspect.isclass(owner):
            if method is None:
                delattr(owner, name)
            else:
                setattr(owner, name, method)
        elif method is None:
            try:
                object.__delattr__(owner, name)
            except TypeError as e:
                if str(e) == "can't apply this __delattr__ to instance object":
                    # This is likely Python 2.6, or early 2.7, where we can't
                    # run object.__delattr__ on old-style classes. We have to
                    # fall back to modifying __dict__. It's not ideal but
                    # doable.
                    del owner.__dict__[name]
        else:
            try:
                object.__setattr__(owner, name, method)
            except TypeError as e:
                if str(e) == "can't apply this __setattr__ to instance object":
                    # Similarly as above, we have to default to dict
                    # manipulation on this version of Python.
                    owner.__dict__[name] = method



from __future__ import unicode_literals

from kgb.agency import SpyAgency
from kgb.contextmanagers import spy_on
from kgb.ops import (SpyOpMatchAny,
                     SpyOpMatchInOrder,
                     SpyOpRaise,
                     SpyOpRaiseInOrder,
                     SpyOpReturn,
                     SpyOpReturnInOrder)


# The version of kgb
#
# This is in the format of:
#
#   (Major, Minor, Micro, alpha/beta/rc/final, Release Number, Released)
#
VERSION = (7, 3, 0, 'final', 0, True)


def get_version_string():
    """Return the kgb version as a human-readable string.

    Returns:
        unicode:
        The kgb version.
    """
    version = '%s.%s' % (VERSION[0], VERSION[1])

    if VERSION[2]:
        version += ".%s" % VERSION[2]

    if VERSION[3] != 'final':
        if VERSION[3] == 'rc':
            version += ' RC%s' % VERSION[4]
        else:
            version += ' %s %s' % (VERSION[3], VERSION[4])

    if not is_release():
        version += " (dev)"

    return version


def get_package_version():
    """Return the kgb version as a Python package version string.

    Returns:
        unicode:
        The kgb package version.
    """
    version = '%s.%s' % (VERSION[0], VERSION[1])

    if VERSION[2]:
        version += '.%s' % VERSION[2]

    tag = VERSION[3]

    if tag != 'final':
        if tag == 'alpha':
            tag = 'a'
        elif tag == 'beta':
            tag = 'b'

        version = '%s%s%s' % (version, tag, VERSION[4])

    return version


def is_release():
    """Return whether this is a released version of kgb.

    Returns:
        bool:
        ``True`` if the version is released. ``False`` if it is still in
        development.
    """
    return VERSION[5]


__version_info__ = VERSION[:-1]
__version__ = get_package_version()


__all__ = [
    '__version__',
    '__version_info__',
    'SpyAgency',
    'SpyOpMatchAny',
    'SpyOpMatchInOrder',
    'SpyOpRaise',
    'SpyOpRaiseInOrder',
    'SpyOpReturn',
    'SpyOpReturnInOrder',
    'VERSION',
    'get_package_version',
    'get_version_string',
    'is_release',
    'spy_on',
]



"""Planned operations for spies to perform."""

from __future__ import unicode_literals

from kgb.errors import UnexpectedCallError


class BaseSpyOperation(object):
    """Base class for a spy operation.

    Spy operations can be performed when a spied-on function is called,
    handling it according to a plan provided by the caller. They're registered
    by passing ``op=`` when spying on the function.

    There are a handful of built-in operations in KGB, but projects can
    subclass this and define their own.
    """

    def handle_call(self, spy_call, *args, **kwargs):
        """Handle a call to this operation.

        Args:
            spy_call (kgb.calls.SpyCall):
                The call to handle.

            *args (tuple):
                Positional arguments passed into the call. This will be
                normalized to not contain an object instance for bound
                method or class methods.

            **kwargs (tuple):
                Keyword arguments passed into the call.

        Returns:
            object:
            The value to return to the caller of the spied function.

        Raises:
            Exception:
                Any exception to raise to the caller of the spied function.
        """
        raise NotImplementedError

    def setup(self, spy, force_unbound=False):
        """Set up the operation.

        This associates the spy with the operation, and then returns a fake
        function to set for the spy, which will in turn call the operation's
        handler.

        Args:
            spy (kgb.spies.FunctionSpy):
                The spy this operation is for.

            force_unbound (bool, optional):
                Whether to force building an unbound fake function. This is
                needed for any functions called by an operation itself (and
                is thus used for nested operations).

                Version Added:
                    6.1

        Returns:
            callable:
            The fake function to set up with the spy.
        """
        self.spy = spy

        if spy.func_type == spy.TYPE_BOUND_METHOD and not force_unbound:
            def fake_func(_self, *args, **kwargs):
                return self._on_spy_call(*args, **kwargs)
        else:
            def fake_func(*args, **kwargs):
                return self._on_spy_call(*args, **kwargs)

        return fake_func

    def _on_spy_call(self, *args, **kwargs):
        """Internal handler for a call to this operation.

        This normalizes and sanity-checks the arguments and then calls
        :py:meth:`handle_call`.

        Args:
            *args (tuple):
                All positional arguments made in the call. This may include the
                object instance for bound methods or the class for
                classmethods.

            **kwargs (dict):
                All keyword arguments made in the call.

        Returns:
            object:
            The value to return to the caller of the spied function.

        Raises:
            Exception:
                Any exception to raise to the caller of the spied function.
        """
        spy = self.spy
        spy_call = spy.last_call

        if spy.func_type == spy.TYPE_UNBOUND_METHOD:
            assert spy_call.called_with(*args[1:], **kwargs)
        else:
            assert spy_call.called_with(*args, **kwargs)

        return self.handle_call(spy_call, *args, **kwargs)


class BaseMatchingSpyOperation(BaseSpyOperation):
    """Base class for a operation that handles calls based on matched rules.

    This helps subclasses to call consumer-defined handlers for calls based on
    some kind of conditions. For instance, based on arguments, or the order in
    which calls are made.
    """

    def __init__(self, calls):
        """Initialize the operation.

        By default, this takes a list of configurations for matching calls,
        which the subclass will use to validate and handle a call.

        The calls are a list of dictionaries with the following keys:

        ``args`` (:py:class:`tuple`, optional):
            Positional arguments for a match.

        ``kwargs`` (:py:class:`dict`, optional):
            Keyword arguments for a match.

        ``call_fake`` (:py:class:`callable`, optional):
            A function to call when all arguments have matched. This takes
            precedence over ``call_original``.

        ``call_original`` (:py:class:`bool`, optional):
            Whether to call the original function. This is the default if
            ``call_fake`` is not provided.

        ``op`` (:py:class:`BaseSpyOperation`, optional):
            A spy operation to call instead of ``call_fake`` or
            ``call_original``.

        Subclasses may define custom keys.

        Version Changed:
            6.1:
            Added support for ``op``.

        Args:
            calls (list of dict):
                A list of call match configurations.
        """
        super(BaseMatchingSpyOperation, self).__init__()

        self._calls = calls

    def setup(self, spy, **kwargs):
        """Set up the operation.

        This invokes the common behavior for setting up a spy operation, and
        then goes through all registered calls to see if any call-provided
        operations also need to be set up.

        Version Added:
            6.1

        Args:
            spy (kgb.spies.FunctionSpy):
                The spy this operation is for.

            **kwargs (dict):
                Additional keyword arguments to pass to the parent.

        Returns:
            callable:
            The fake function to set up with the spy.
        """
        result = super(BaseMatchingSpyOperation, self).setup(spy, **kwargs)

        new_calls = []

        # Convert any operations into fake functions.
        #
        # Note that we have to force the resulting fake function to be
        # unbound, since all fake functions provided to an operation are
        # called without an owner.
        for call in self._calls:
            if 'op' in call:
                call = call.copy()
                call['call_fake'] = call.pop('op').setup(spy,
                                                         force_unbound=True)

            new_calls.append(call)

        self._calls = new_calls

        return result

    def get_call_match_config(self, spy_call):
        """Return a call match configuration for a call.

        This will typically be one of the call match configurations provided
        during initialization.

        Subclasses must override this to return a dictionary containing
        information that can be used to assert, track, and handle a call.
        If they can't find a suitable call, they must raise
        :py:class:`kgb.errors.UnexpectedCallError`.

        Args:
            spy_call (kgb.calls.SpyCall):
                The call to return a match for.

        Returns:
            dict:
            The call match configuration.

        Raises:
            kgb.errors.UnexpectedCallError:
                A call match configuration could not be found. Details should
                be in the error message.
        """
        raise NotImplementedError

    def validate_call(self, call_match_config):
        """Validate that the last call matches the call configuration.

        This will assert that the last call matches the ``args`` and ``kwargs``
        from the given call match configuration.

        Subclasses can override this to check other conditions.

        Args:
            call_match_config (dict):
                The call match configuration returned from
                :py:meth:`get_call_match_config` for the last call.

        Raises:
            AssertionError:
                The call did not match the configuration.
        """
        self.spy.agency.assertSpyCalledWith(
            self.spy.last_call,
            *call_match_config.get('args', ()),
            **call_match_config.get('kwargs', {}))

    def handle_call(self, spy_call, *args, **kwargs):
        """Handle a call to this operation.

        This will find a suitable call match configuration, if one was
        provided, and then call one of the following, in order of preference:

        1. The spy operation (if ``op`` is set)
        2. The fake function (if ``call_fake`` was provided)
        3. The original function (if ``call_original`` is not set to
           ``False``)

        If none of the above are invoked, ``None`` will be returned instead.

        Version Changed:
            6.1:
            Added support for ``op``.

        Args:
            spy_call (kgb.calls.SpyCall):
                The call to handle.

            *args (tuple):
                Positional arguments passed into the call. This will be
                normalized to not contain an object instance for bound
                method or class methods.

            **kwargs (tuple):
                Keyword arguments passed into the call.

        Returns:
            object:
            The value to return to the caller of the spied function.
            This may be returned by a fake or original function.

        Raises:
            AssertionError:
                The call did not match the returned configuration.

            Exception:
                Any exception to raise to the caller of the spied function.
                This may be raised by a fake or original function.

            kgb.errors.UnexpectedCallError:
                A call match configuration could not be found. Details should
                be in the error message.
        """
        call_match_config = self.get_call_match_config(spy_call)
        assert call_match_config is not None

        self.validate_call(call_match_config)

        # We'll be respecting these arguments in the order that FunctionSpy
        # would with its parameters.
        func = call_match_config.get('call_fake')

        if func is not None:
            return func(*args, **kwargs)

        if call_match_config.get('call_original', True):
            return self.spy.call_original(*args, **kwargs)

        return None


class SpyOpMatchAny(BaseMatchingSpyOperation):
    """A operation for handling one or more expected calls in any order.

    This is used to list the calls (specifying positional and keyword
    arguments) that are expected to be made, raising an error if any calls are
    made that weren't expected.

    Each of those expected sets of arguments can optionally result in a call to
    a fake function or the original function. This can be specified per set of
    arguments.

    Example:
        spy_on(traps.trigger, op=SpyOpMatchAny([
            {
                'args': ('hallway_lasers',),
                'call_fake': _send_wolves,
            },
            {
                'args': ('trap_tile',),
                'call_fake': _spill_hot_oil,
            },
            {
                'args': ('infrared_camera',),
                'kwargs': {
                    'sector': 'underground_passage',
                },
                'call_original': False,
            },
        ]))
    """

    def __init__(self, calls):
        """Initialize the operation.

        This takes a list of configurations for matching calls, which can be
        called in any order.
        those calls are expected.

        The calls are a list of dictionaries with the following keys:

        ``args`` (:py:class:`tuple`, optional):
            Positional arguments for a match.

        ``kwargs`` (:py:class:`dict`, optional):
            Keyword arguments for a match.

        ``call_fake`` (:py:class:`callable`, optional):
            A function to call when all arguments have matched. This takes
            precedence over ``call_original``.

        ``call_original`` (:py:class:`bool`, optional):
            Whether to call the original function. This is the default if
            ``call_fake`` is not provided.

        Args:
            calls (list of dict):
                A list of call match configurations.
        """
        super(SpyOpMatchAny, self).__init__(calls)

    def get_call_match_config(self, spy_call):
        """Return a call match configuration for a call.

        This will check if there are any call match configurations provided
        during initialization that match the call.

        Args:
            spy_call (kgb.calls.SpyCall):
                The call to return a match for.

        Returns:
            dict:
            The call match configuration.

        Raises:
            kgb.errors.UnexpectedCallError:
                A call match configuration could not be found. Details should
                be in the error message.
        """
        for call_match_config in self._calls:
            if spy_call.called_with(*call_match_config.get('args', ()),
                                    **call_match_config.get('kwargs', {})):
                return call_match_config

        raise UnexpectedCallError(
            '%(spy)s was not called with any expected arguments.'
            % {
                'spy': self.spy.func_name,
            })


class SpyOpMatchInOrder(BaseMatchingSpyOperation):
    """A operation for handling expected calls in a given order.

    This is used to list the calls (specifying positional and keyword
    arguments) that are expected to be made, in the order they should be made,
    raising an error if too many calls were made or a call didn't match the
    expected arguments.

    Each of those expected sets of arguments can optionally result in a call to
    a fake function or the original function. This can be specified per set of
    arguments.

    Example:
        spy_on(lockbox.enter_code, op=SpyOpMatchInOrder([
            {
                'args': (1, 2, 3, 4, 5, 6),
                'call_original': False,
            },
            {
                'args': (9, 0, 2, 1, 0, 0),
                'call_fake': _start_countdown,
            },
            {
                'args': (4, 8, 15, 16, 23, 42),
                'kwargs': {
                    'secret_button_pushed': True,
                },
                'call_original': True,
            },
            {
                'args': (4, 8, 15, 16, 23, 42),
                'kwargs': {
                    'secret_button_pushed': True,
                },
                'op': SpyOpRaise(Exception('Oh no')),
            },
        ]))
    """

    def __init__(self, calls):
        """Initialize the operation.

        This takes a list of configurations for matching calls, in the order
        those calls are expected.

        The calls are a list of dictionaries with the following keys:

        ``args`` (:py:class:`tuple`, optional):
            Positional arguments for a match.

        ``kwargs`` (:py:class:`dict`, optional):
            Keyword arguments for a match.

        ``call_fake`` (:py:class:`callable`, optional):
            A function to call when all arguments have matched. This takes
            precedence over ``call_original``.

        ``call_original`` (:py:class:`bool`, optional):
            Whether to call the original function. This is the default if
            ``call_fake`` is not provided.

        ``op`` (:py:class:`BaseSpyOperation`, optional):
            A spy operation to call instead of ``call_fake`` or
            ``call_original``.

        Args:
            calls (list of dict):
                A list of call match configurations.
        """
        super(SpyOpMatchInOrder, self).__init__(calls)

        self._next = 0

    def get_call_match_config(self, spy_call):
        """Return a call match configuration for a call.

        This will check if the spy call matches the next call match
        configuration in the list provided by the consumer.

        Args:
            spy_call (kgb.calls.SpyCall):
                The call to return a match for.

        Returns:
            dict:
            The call match configuration.

        Raises:
            kgb.errors.UnexpectedCallError:
                Too many calls were made to the function.
        """
        i = self._next

        try:
            call_match_config = self._calls[i]
        except IndexError:
            raise UnexpectedCallError(
                '%(spy)s was called %(num_calls)s time(s), but only '
                '%(expected_calls)s call(s) were expected. Latest call: '
                '%(latest_call)s'
                % {
                    'expected_calls': len(self._calls),
                    'latest_call': spy_call,
                    'num_calls': i + 1,
                    'spy': self.spy.func_name,
                })

        self._next += 1

        return call_match_config


class SpyOpRaise(BaseSpyOperation):
    """An operation for raising an exception.

    This is used to simulate a failure of some sort in a function or method.

    Example:
        spy_on(pen.emit_poison, op=SpyOpRaise(PoisonEmptyError()))
    """

    def __init__(self, exc):
        """Initialize the operation.

        Args:
            exc (Exception):
                The exception instance to raise when the function is called.
        """
        self.exc = exc

    def handle_call(self, *args, **kwargs):
        """Handle a call to this operation.

        This will raise the exception provided to the operation.

        Args:
            *args (tuple, ignored):
                Positional arguments passed into the call.

            **kwargs (tuple, ignored):
                Keyword arguments passed into the call.

        Raises:
            Exception:
                The exception provided to the operation.
        """
        raise self.exc


class SpyOpRaiseInOrder(SpyOpMatchInOrder):
    """An operation for raising exceptions in the order of calls.

    This is similar to :py:class:`SpyOpRaise`, but will raise a different
    exception for each call, based on a provided list.

    Example:
        spy_on(our_agent.get_identities, op=SpyOpRaiseInOrder([
            PoisonEmptyError(),
            Kaboom(),
            MissingPenError(),
        ]))
    """

    def __init__(self, exceptions):
        """Initialize the operation.

        Args:
            exceptions (list of Exception):
                The list of exceptions, one for each function call.
        """
        super(SpyOpRaiseInOrder, self).__init__([
            {
                'op': SpyOpRaise(exc),
            }
            for exc in exceptions
        ])


class SpyOpReturn(BaseSpyOperation):
    """An operation for returning a value.

    This is used to simulate a simple result from a function call without
    having to override the method or provide a lambda.

    Example:
        spy_on(our_agent.get_identity, op=SpyOpReturn('nobody...'))
    """

    def __init__(self, return_value):
        """Initialize the operation.

        Args:
            return_value (object):
                The value to return when the function is called.
        """
        self.return_value = return_value

    def handle_call(self, *args, **kwargs):
        """Handle a call to this operation.

        This will return the value provided to the operation.

        Args:
            *args (tuple, ignored):
                Positional arguments passed into the call.

            **kwargs (tuple, ignored):
                Keyword arguments passed into the call.

        Returns:
            object:
            The return value provided to the operation.
        """
        return self.return_value


class SpyOpReturnInOrder(SpyOpMatchInOrder):
    """An operation for returning a value.

    This is similar to :py:class:`SpyOpReturn`, but will return a different
    value for each call, based on a provided list.

    Example:
        spy_on(our_agent.get_identities, op=SpyOpReturnInOrder([
            'nobody...',
            'who?',
            'never heard of them...',
        ]))
    """

    def __init__(self, return_values):
        """Initialize the operation.

        Args:
            return_values (list):
                The list of values, one for each function call.
        """
        super(SpyOpReturnInOrder, self).__init__([
            {
                'op': SpyOpReturn(value),
            }
            for value in return_values
        ])



"""Unit tests for kgb.pytest_plugin.

Version Added:
    7.0
"""

from __future__ import unicode_literals

from kgb.agency import SpyAgency
from kgb.tests.base import MathClass


def test_pytest_plugin(spy_agency):
    """Testing pytest spy_agency fixture"""
    assert isinstance(spy_agency, SpyAgency)

    obj = MathClass()

    spy = spy_agency.spy_on(obj.do_math)
    assert spy_agency.spies == {spy}



#!/usr/bin/env python
from __future__ import print_function, unicode_literals

import os
import sys

import pytest


if __name__ == '__main__':
    print('This is deprecated! Please run pytest instead.')
    print()
    os.chdir(os.path.join(os.path.dirname(__file__), '..'))
    sys.exit(pytest.main(sys.argv[1:]))



import functools
import inspect
import logging
import sys
from unittest import SkipTest

from kgb.tests.base import TestCase


logger = logging.getLogger('kgb')


def require_func_pos_only_args(func):
    """Require positional-only arguments for a function.

    If not available, the test will be skippd.

    Args:
        func (callable):
            The unit test function to decorate.
    """
    @functools.wraps(func)
    def _wrap(*args, **kwargs):
        if sys.version_info[:2] >= (3, 8):
            return func(*args, **kwargs)
        else:
            raise SkipTest('inspect.getargspec is not available on Python 3.8')

    return _wrap


class FunctionSpyTests(TestCase):
    """Python 3 unit tests for kgb.spies.FunctionSpy."""

    def test_spy_with_function_copies_attribute_state(self):
        """Testing FunctionSpy with functions copies all attribute states"""
        def func(a: str, b: int = 1, *, c: int = 3) -> bool:
            return False

        self.agency.spy_on(func)
        spy_func = func.spy.func

        self.assertIsNot(spy_func, func)
        self.assertEqual(
            spy_func.__kwdefaults__,
            {
                'c': 3,
            })
        self.assertEqual(
            spy_func.__annotations__,
            {
                'a': str,
                'b': int,
                'c': int,
                'return': bool,
            })

    def test_spy_with_bound_methods_copies_attribute_state(self):
        """Testing FunctionSpy with bound methods copies all attribute states
        """
        class A:
            def func(a: str, b: int = 1, *, c: int = 3) -> bool:
                return False

        a = A()
        self.agency.spy_on(a.func)
        spy_func = a.func.spy.func

        self.assertIsNot(spy_func, a.func)
        self.assertEqual(
            spy_func.__kwdefaults__,
            {
                'c': 3,
            })
        self.assertEqual(
            spy_func.__annotations__,
            {
                'a': str,
                'b': int,
                'c': int,
                'return': bool,
            })

    def test_spy_with_unbound_methods_copies_attribute_state(self):
        """Testing FunctionSpy with unbound methods copies all attribute states
        """
        class A:
            def func(a: str, b: int = 1, *, c: int = 3) -> bool:
                return False

        self.agency.spy_on(A.func)
        spy_func = A.func.spy.func

        self.assertIsNot(spy_func, A.func)
        self.assertEqual(
            spy_func.__kwdefaults__,
            {
                'c': 3,
            })
        self.assertEqual(
            spy_func.__annotations__,
            {
                'a': str,
                'b': int,
                'c': int,
                'return': bool,
            })

    @require_func_pos_only_args
    def test_call_with_function_and_positional_only_args(self):
        """Testing FunctionSpy calls with function containing positional-only
        arguments
        """
        func = self.make_func("""
            def func(a, b=1, /):
                return a * b
        """)

        self.agency.spy_on(func)
        result = func(2, 5)

        self.assertEqual(result, 10)
        self.assertEqual(len(func.spy.calls), 1)
        self.assertEqual(func.spy.calls[0].args, (2, 5))
        self.assertEqual(func.spy.calls[0].kwargs, {})

    @require_func_pos_only_args
    def test_call_with_function_and_positional_only_args_no_pos_passed(self):
        """Testing FunctionSpy calls with function containing positional-only
        arguments and no positional argument passed
        """
        func = self.make_func("""
            def func(a, b=2, /):
                return a * b
        """)

        self.agency.spy_on(func)
        result = func(2)

        self.assertEqual(result, 4)
        self.assertEqual(len(func.spy.calls), 1)
        self.assertEqual(func.spy.calls[0].args, (2, 2))
        self.assertEqual(func.spy.calls[0].kwargs, {})

    def test_call_with_function_and_keyword_only_args(self):
        """Testing FunctionSpy calls with function containing keyword-only
        arguments
        """
        def func(a, *, b=2):
            return a * b

        self.agency.spy_on(func)
        result = func(2, b=5)

        self.assertEqual(result, 10)
        self.assertEqual(len(func.spy.calls), 1)
        self.assertEqual(func.spy.calls[0].args, (2,))
        self.assertEqual(func.spy.calls[0].kwargs, {'b': 5})

    def test_call_with_function_and_keyword_only_args_no_kw_passed(self):
        """Testing FunctionSpy calls with function containing keyword-only
        arguments and no keyword passed
        """
        def func(a, *, b=2):
            return a * b

        self.agency.spy_on(func)
        result = func(2)

        self.assertEqual(result, 4)
        self.assertEqual(len(func.spy.calls), 1)
        self.assertEqual(func.spy.calls[0].args, (2,))
        self.assertEqual(func.spy.calls[0].kwargs, {'b': 2})

    def test_init_with_unbound_method_decorator_bad_func_name(self):
        """Testing FunctionSpy construction with a decorator not preserving
        an unbound method name
        """
        def bad_deco(func):
            def _wrapper(*args, **kwargs):
                return func(*args, **kwargs)

            return _wrapper

        class MyObject(object):
            @bad_deco
            def my_method(self):
                pass

        self.agency.spy_on(logger.warning)

        self.agency.spy_on(MyObject.my_method,
                           owner=MyObject)

        self.agency.assertSpyCalledWith(
            logger.warning,
            "%r doesn't have a function named \"%s\". This "
            "appears to be a decorator that doesn't "
            "preserve function names. Try passing "
            "func_name= when setting up the spy.",
            MyObject,
            '_wrapper')

    def test_init_with_unbound_method_decorator_corrected_func_name(self):
        """Testing FunctionSpy construction with a decorator not preserving
        an unbound method name and explicit func_name= provided
        """
        def bad_deco(func):
            def _wrapper(*args, **kwargs):
                return func(*args, **kwargs)

            return _wrapper

        class MyObject(object):
            @bad_deco
            def my_method(self):
                pass

        self.agency.spy_on(logger.warning)

        self.agency.spy_on(MyObject.my_method,
                           owner=MyObject,
                           func_name='my_method')

        self.agency.assertSpyNotCalled(logger.warning)

    def test_getfullargspec_with_function(self):
        """Testing FunctionSpy in inspect.getfullargspec() with function"""
        def func(a, b=2):
            return a * b

        self.agency.spy_on(func)

        argspec = inspect.getfullargspec(func)

        self.assertEqual(argspec.args, ['a', 'b'])
        self.assertIsNone(argspec.varargs)
        self.assertIsNone(argspec.varkw)
        self.assertEqual(argspec.defaults, (2,))
        self.assertEqual(argspec.kwonlyargs, [])
        self.assertIsNone(argspec.kwonlydefaults)
        self.assertEqual(argspec.annotations, {})

    @require_func_pos_only_args
    def test_getfullargspec_with_function_pos_only(self):
        """Testing FunctionSpy in inspect.getfullargspec() with function and
        positional-only arguments
        """
        func = self.make_func("""
            def func(a, b=2, /, c=3):
                return a * b
        """)

        self.agency.spy_on(func)

        argspec = inspect.getfullargspec(func)

        self.assertEqual(argspec.args, ['a', 'b', 'c'])
        self.assertIsNone(argspec.varargs)
        self.assertIsNone(argspec.varkw)
        self.assertEqual(argspec.defaults, (2, 3))
        self.assertEqual(argspec.kwonlyargs, [])
        self.assertIsNone(argspec.kwonlydefaults)
        self.assertEqual(argspec.annotations, {})

    def test_getfullargspec_with_function_keyword_only(self):
        """Testing FunctionSpy in inspect.getfullargspec() with function and
        keyword-only arguments
        """
        def func(*, a, b=2):
            return a * b

        self.agency.spy_on(func)

        argspec = inspect.getfullargspec(func)

        self.assertEqual(argspec.args, [])
        self.assertIsNone(argspec.varargs)
        self.assertIsNone(argspec.varkw)
        self.assertIsNone(argspec.defaults)
        self.assertEqual(argspec.kwonlyargs, ['a', 'b'])
        self.assertEqual(argspec.kwonlydefaults, {
            'b': 2,
        })
        self.assertEqual(argspec.annotations, {})

    def test_getfullargspec_with_function_annotations(self):
        """Testing FunctionSpy in inspect.getfullargspec() with function and
        annotations
        """
        def func(a: int, b: int = 2) -> int:
            return a * b

        self.agency.spy_on(func)

        argspec = inspect.getfullargspec(func)

        self.assertEqual(argspec.args, ['a', 'b'])
        self.assertIsNone(argspec.varargs)
        self.assertIsNone(argspec.varkw)
        self.assertEqual(argspec.defaults, (2,))
        self.assertEqual(argspec.kwonlyargs, [])
        self.assertIsNone(argspec.kwonlydefaults)
        self.assertEqual(argspec.annotations, {
            'a': int,
            'b': int,
            'return': int,
        })

    def test_getfullargspec_with_bound_method(self):
        """Testing FunctionSpy in inspect.getfullargspec() with bound method"""
        class MyObject:
            def func(self, a, b=2):
                return a * b

        obj = MyObject()
        self.agency.spy_on(obj.func)

        argspec = inspect.getfullargspec(obj.func)

        self.assertEqual(argspec.args, ['self', 'a', 'b'])
        self.assertIsNone(argspec.varargs)
        self.assertIsNone(argspec.varkw)
        self.assertEqual(argspec.defaults, (2,))
        self.assertEqual(argspec.kwonlyargs, [])
        self.assertIsNone(argspec.kwonlydefaults)
        self.assertEqual(argspec.annotations, {})

    @require_func_pos_only_args
    def test_getfullargspec_with_bound_method_pos_only(self):
        """Testing FunctionSpy in inspect.getfullargspec() with bound method
        and positional-only arguments
        """
        MyObject = self.make_func(
            """
            class MyObject:
                def func(self, a, b=2, /, c=3):
                    return a * b
            """,
            func_name='MyObject')

        obj = MyObject()
        self.agency.spy_on(obj.func)

        argspec = inspect.getfullargspec(obj.func)

        self.assertEqual(argspec.args, ['self', 'a', 'b', 'c'])
        self.assertIsNone(argspec.varargs)
        self.assertIsNone(argspec.varkw)
        self.assertEqual(argspec.defaults, (2, 3))
        self.assertEqual(argspec.kwonlyargs, [])
        self.assertIsNone(argspec.kwonlydefaults)
        self.assertEqual(argspec.annotations, {})

    def test_getfullargspec_with_bound_method_keyword_only(self):
        """Testing FunctionSpy in inspect.getfullargspec() with bound method
        and keyword-only arguments
        """
        class MyObject:
            def func(self, *, a, b=2):
                return a * b

        obj = MyObject()
        self.agency.spy_on(obj.func)

        argspec = inspect.getfullargspec(obj.func)

        self.assertEqual(argspec.args, ['self'])
        self.assertIsNone(argspec.varargs)
        self.assertIsNone(argspec.varkw)
        self.assertIsNone(argspec.defaults)
        self.assertEqual(argspec.kwonlyargs, ['a', 'b'])
        self.assertEqual(argspec.kwonlydefaults, {
            'b': 2,
        })
        self.assertEqual(argspec.annotations, {})

    def test_getfullargspec_with_bound_method_annotations(self):
        """Testing FunctionSpy in inspect.getfullargspec() with bound method
        and annotations
        """
        class MyObject:
            def func(self, a: int, b: int = 2) -> int:
                return a * b

        obj = MyObject()
        self.agency.spy_on(obj.func)

        argspec = inspect.getfullargspec(obj.func)

        self.assertEqual(argspec.args, ['self', 'a', 'b'])
        self.assertIsNone(argspec.varargs)
        self.assertIsNone(argspec.varkw)
        self.assertEqual(argspec.defaults, (2,))
        self.assertEqual(argspec.kwonlyargs, [])
        self.assertIsNone(argspec.kwonlydefaults)
        self.assertEqual(argspec.annotations, {
            'a': int,
            'b': int,
            'return': int,
        })

    def test_getfullargspec_with_unbound_method(self):
        """Testing FunctionSpy in inspect.getfullargspec() with unbound method
        """
        class MyObject:
            def func(self, a, b=2):
                return a * b

        self.agency.spy_on(MyObject.func)

        argspec = inspect.getfullargspec(MyObject.func)

        self.assertEqual(argspec.args, ['self', 'a', 'b'])
        self.assertIsNone(argspec.varargs)
        self.assertIsNone(argspec.varkw)
        self.assertEqual(argspec.defaults, (2,))
        self.assertEqual(argspec.kwonlyargs, [])
        self.assertIsNone(argspec.kwonlydefaults)
        self.assertEqual(argspec.annotations, {})

    @require_func_pos_only_args
    def test_getfullargspec_with_unbound_method_pos_only(self):
        """Testing FunctionSpy in inspect.getfullargspec() with unbound method
        and positional-only arguments
        """
        MyObject = self.make_func(
            """
            class MyObject:
                def func(self, a, b=2, /, c=3):
                    return a * b
            """,
            func_name='MyObject')

        self.agency.spy_on(MyObject.func)

        argspec = inspect.getfullargspec(MyObject.func)

        self.assertEqual(argspec.args, ['self', 'a', 'b', 'c'])
        self.assertIsNone(argspec.varargs)
        self.assertIsNone(argspec.varkw)
        self.assertEqual(argspec.defaults, (2, 3))
        self.assertEqual(argspec.kwonlyargs, [])
        self.assertIsNone(argspec.kwonlydefaults)
        self.assertEqual(argspec.annotations, {})

    def test_getfullargspec_with_unbound_method_keyword_only(self):
        """Testing FunctionSpy in inspect.getfullargspec() with unbound method
        and keyword-only arguments
        """
        class MyObject:
            def func(self, *, a, b=2):
                return a * b

        self.agency.spy_on(MyObject.func)

        argspec = inspect.getfullargspec(MyObject.func)

        self.assertEqual(argspec.args, ['self'])
        self.assertIsNone(argspec.varargs)
        self.assertIsNone(argspec.varkw)
        self.assertIsNone(argspec.defaults)
        self.assertEqual(argspec.kwonlyargs, ['a', 'b'])
        self.assertEqual(argspec.kwonlydefaults, {
            'b': 2,
        })
        self.assertEqual(argspec.annotations, {})

    def test_getfullargspec_with_unbound_method_annotations(self):
        """Testing FunctionSpy in inspect.getfullargspec() with unbound method
        and annotations
        """
        class MyObject:
            def func(self, a: int, b: int = 2) -> int:
                return a * b

        self.agency.spy_on(MyObject.func)

        argspec = inspect.getfullargspec(MyObject.func)

        self.assertEqual(argspec.args, ['self', 'a', 'b'])
        self.assertIsNone(argspec.varargs)
        self.assertIsNone(argspec.varkw)
        self.assertEqual(argspec.defaults, (2,))
        self.assertEqual(argspec.kwonlyargs, [])
        self.assertIsNone(argspec.kwonlydefaults)
        self.assertEqual(argspec.annotations, {
            'a': int,
            'b': int,
            'return': int,
        })

    def test_getfullargspec_with_classmethod(self):
        """Testing FunctionSpy in inspect.getfullargspec() with classmethod
        """
        class MyObject:
            @classmethod
            def func(cls, a, b=2):
                return a * b

        self.agency.spy_on(MyObject.func)

        argspec = inspect.getfullargspec(MyObject.func)

        self.assertEqual(argspec.args, ['cls', 'a', 'b'])
        self.assertIsNone(argspec.varargs)
        self.assertIsNone(argspec.varkw)
        self.assertEqual(argspec.defaults, (2,))
        self.assertEqual(argspec.kwonlyargs, [])
        self.assertIsNone(argspec.kwonlydefaults)
        self.assertEqual(argspec.annotations, {})

    @require_func_pos_only_args
    def test_getfullargspec_with_classmethod_pos_only(self):
        """Testing FunctionSpy in inspect.getfullargspec() with classmethod
        and positional-only arguments
        """
        MyObject = self.make_func(
            """
            class MyObject:
                @classmethod
                def func(cls, a, b=2, /, c=3):
                    return a * b
            """,
            func_name='MyObject')

        self.agency.spy_on(MyObject.func)

        argspec = inspect.getfullargspec(MyObject.func)

        self.assertEqual(argspec.args, ['cls', 'a', 'b', 'c'])
        self.assertIsNone(argspec.varargs)
        self.assertIsNone(argspec.varkw)
        self.assertEqual(argspec.defaults, (2, 3))
        self.assertEqual(argspec.kwonlyargs, [])
        self.assertIsNone(argspec.kwonlydefaults)
        self.assertEqual(argspec.annotations, {})

    def test_getfullargspec_with_classmethod_keyword_only(self):
        """Testing FunctionSpy in inspect.getfullargspec() with classmethod
        and keyword-only arguments
        """
        class MyObject:
            @classmethod
            def func(cls, *, a, b=2):
                return a * b

        self.agency.spy_on(MyObject.func)

        argspec = inspect.getfullargspec(MyObject.func)

        self.assertEqual(argspec.args, ['cls'])
        self.assertIsNone(argspec.varargs)
        self.assertIsNone(argspec.varkw)
        self.assertIsNone(argspec.defaults)
        self.assertEqual(argspec.kwonlyargs, ['a', 'b'])
        self.assertEqual(argspec.kwonlydefaults, {
            'b': 2,
        })
        self.assertEqual(argspec.annotations, {})

    def test_getfullargspec_with_classmethod_annotations(self):
        """Testing FunctionSpy in inspect.getfullargspec() with classmethod
        and annotations
        """
        class MyObject:
            @classmethod
            def func(cls, a: int, b: int = 2) -> int:
                return a * b

        self.agency.spy_on(MyObject.func)

        argspec = inspect.getfullargspec(MyObject.func)

        self.assertEqual(argspec.args, ['cls', 'a', 'b'])
        self.assertIsNone(argspec.varargs)
        self.assertIsNone(argspec.varkw)
        self.assertEqual(argspec.defaults, (2,))
        self.assertEqual(argspec.kwonlyargs, [])
        self.assertIsNone(argspec.kwonlydefaults)
        self.assertEqual(argspec.annotations, {
            'a': int,
            'b': int,
            'return': int,
        })



"""A spy agency to manage spies."""

from __future__ import unicode_literals

from pprint import pformat
from unittest.util import safe_repr

from kgb.signature import _UNSET_ARG
from kgb.spies import FunctionSpy, SpyCall
from kgb.utils import format_spy_kwargs


class SpyAgency(object):
    """Manages spies.

    A SpyAgency can be instantiated or mixed into a
    :py:class:`unittest.TestCase` in order to provide spies.

    Every spy created through this agency will be tracked, and can be later
    be removed (individually or at once).

    Version Changed:
        7.0:
        Added ``assert_`` versions of all the assertion methods (e.g.,
        ``assert_spy_called_with`` as an alias of ``assertSpyCalledWith``.

    Attributes:
        spies (set of kgb.spies.FunctionSpy):
            All spies currently registered with this agency.
    """

    def __init__(self, *args, **kwargs):
        """Initialize the spy agency.

        Args:
            *args (tuple):
                Positional arguments to pass on to any other class (if using
                this as a mixin).

            **kwargs (dict):
                Keyword arguments to pass on to any other class (if using
                this as a mixin).
        """
        super(SpyAgency, self).__init__(*args, **kwargs)

        self.spies = set()

    def tearDown(self):
        """Tear down a test suite.

        This is used when SpyAgency is mixed into a TestCase. It will
        automatically remove all spies when tearing down.
        """
        super(SpyAgency, self).tearDown()
        self.unspy_all()

    def spy_on(self, *args, **kwargs):
        """Spy on a function.

        By default, the spy will allow the call to go through to the original
        function. This can be disabled by passing ``call_original=False`` when
        initiating the spy. If disabled, the original function will never be
        called.

        This can also be passed a ``call_fake`` parameter pointing to another
        function to call instead of the original. If passed, this will take
        precedence over ``call_original``.

        See :py:class:`~kgb.spies.FunctionSpy` for more details on arguments.

        Args:
            *args (tuple):
                Positional arguments to pass to
                :py:class:`~kgb.spies.FunctionSpy`.

            **kwargs (dict):
                Keyword arguments to pass to
                :py:class:`~kgb.spies.FunctionSpy`.

        Returns:
            kgb.spies.FunctionSpy:
            The resulting spy.
        """
        spy = FunctionSpy(self, *args, **kwargs)
        self.spies.add(spy)
        return spy

    def spy_for(self, func, owner=_UNSET_ARG):
        """Decorate a function that should be a spy for another function.

        This is a convenience over declaring a function and using
        :py:meth:`spy_on` with ``call_fake=``. It's used to quickly and
        easily create a fake function spy for another function.

        Version Added:
            6.0

        Args:
            func (callable):
                The function or method to spy on.

            owner (type or object, optional):
                The owner of the function or method.

                If spying on an unbound method, this **must** be set to the
                class that owns it.

                If spying on a bound method that identifies as a plain
                function (which may happen if the method is decorated and
                dynamically returns a new function on access), this should
                be the instance of the object you're spying on.

        Example:
            @self.spy_for(get_doomsday):
            def _fake_get_doomsday():
                return datetime(year=2038, month=12, day=5,
                                hour=1, minute=2, second=3)
        """
        def _wrap(call_fake):
            self.spy_on(func,
                        owner=owner,
                        call_fake=call_fake)

            return call_fake

        return _wrap

    def unspy(self, func):
        """Stop spying on a function.

        Args:
            func (callable):
                The function to stop spying on.

        Raises:
            ValueError:
                The provided function was not spied on.
        """
        try:
            spy = func.spy
        except AttributeError:
            raise ValueError('Function %r has not been spied on.' % func)

        assert spy in self.spies

        spy.unspy()

    def unspy_all(self):
        """Stop spying on all functions tracked by this agency."""
        for spy in self.spies:
            spy.unspy(unregister=False)

        self.spies.clear()

    def get_spy(self, func):
        """Return the spy for a function.

        This method is a helper to prevent type checkers from complaining when
        accessing data about a spy.

        Version Added:
            7.3

        Args:
            func (callable):
                The function to return the spy for.

        Returns:
            kgb.spies.FunctionSpy:
            The function's spy.

        Raises:
            ValueError:
                The provided function was not spied on.
        """
        try:
            return func.spy
        except AttributeError:
            raise ValueError('Function %r has not been spied on.' % func)

    def assertHasSpy(self, spy):
        """Assert that a function has a spy.

        This also accepts a spy as an argument, which will always return
        ``True``.

        Args:
            spy (callable or kgb.spies.FunctionSpy):
                The function or spy to check.

        Raises:
            AssertionError:
                The function did not have a spy.
        """
        if not hasattr(spy, 'spy') and not isinstance(spy, FunctionSpy):
            self._kgb_assert_fail('%s has not been spied on.'
                                  % self._format_spy_or_call(spy))

    def assertSpyCalled(self, spy):
        """Assert that a function has been called at least once.

        This will imply :py:meth:`assertHasSpy`.

        Args:
            spy (callable or kgb.spies.FunctionSpy):
                The function or spy to check.

        Raises:
            AssertionError:
                The function was not called.
        """
        self.assertHasSpy(spy)

        if not spy.called:
            self._kgb_assert_fail('%s was not called.'
                                  % self._format_spy_or_call(spy))

    def assertSpyCalledOnce(self, spy):
        """Assert that a function has been called exactly once.

        This will imply :py:meth:`assertHasSpy`.

        Args:
            spy (callable or kgb.spies.FunctionSpy):
                The function or spy to check.

        Raises:
            AssertionError:
                The function was not called.
        """
        self.assertSpyCalled(spy)

        # The spy's been called at least once. If it's called more than once,
        # show the calls.
        call_count = len(spy.calls)

        if call_count > 1:
            self._kgb_assert_fail(
                '%s was not called exactly 1 time. It was called %d times:\n'
                '\n'
                '%s'
                % (
                    self._format_spy_or_call(spy),
                    call_count,
                    self._format_spy_calls(spy, self._format_spy_call_args),
                ))

    def assertSpyNotCalled(self, spy):
        """Assert that a function has not been called.

        This will imply :py:meth:`assertHasSpy`.

        Args:
            spy (callable or kgb.spies.FunctionSpy):
                The function or spy to check.

        Raises:
            AssertionError:
                The function was called.
        """
        self.assertHasSpy(spy)

        if spy.called:
            call_count = len(spy.calls)

            if call_count == 1:
                msg = (
                    '%s was called 1 time:'
                    % self._format_spy_or_call(spy)
                )
            else:
                msg = (
                    '%s was called %d times:'
                    % (self._format_spy_or_call(spy), call_count)
                )

            self._kgb_assert_fail(
                '%s\n'
                '\n'
                '%s'
                % (
                    msg,
                    self._format_spy_calls(spy, self._format_spy_call_args),
                ))

    def assertSpyCallCount(self, spy, count):
        """Assert that a function was called the given number of times.

        This will imply :py:meth:`assertHasSpy`.

        Args:
            spy (callable or kgb.spies.FunctionSpy):
                The function or spy to check.

            count (int):
                The number of times the function is expected to have been
                called.

        Raises:
            AssertionError:
                The function was not called the specified number of times.
        """
        self.assertHasSpy(spy)

        call_count = len(spy.calls)

        if call_count != count:
            if call_count == 1:
                msg = '%s was called %d time, not %d.'
            else:
                msg = '%s was called %d times, not %d.'

            self._kgb_assert_fail(msg %
                                  (self._format_spy_or_call(spy),
                                   call_count,
                                   count))

    def assertSpyCalledWith(self, spy_or_call, *expected_args,
                            **expected_kwargs):
        """Assert that a function was called with the given arguments.

        If a spy is provided, all calls will be checked for a match.

        This will imply :py:meth:`assertHasSpy`.

        Args:
            spy_or_call (callable or kgb.spies.FunctionSpy):
                The function, spy, or call to check.

            *expected_args (tuple):
                Positional arguments expected to be provided in any of the
                calls.

            **expected_kwargs (dict):
                Keyword arguments expected to be provided in any of the calls.

        Raises:
            AssertionError:
                The function was not called with the provided arguments.
        """
        if isinstance(spy_or_call, FunctionSpy):
            self.assertSpyCalled(spy_or_call)

        if not spy_or_call.called_with(*expected_args, **expected_kwargs):
            if isinstance(spy_or_call, SpyCall):
                self._kgb_assert_fail(
                    'This call to %s was not passed args=%s, kwargs=%s.\n'
                    '\n'
                    'It was called with:\n'
                    '\n'
                    '%s'
                    % (
                        self._format_spy_or_call(spy_or_call),
                        safe_repr(expected_args),
                        format_spy_kwargs(expected_kwargs),
                        self._format_spy_call_args(spy_or_call),
                    ))
            else:
                self._kgb_assert_fail(
                    'No call to %s was passed args=%s, kwargs=%s.\n'
                    '\n'
                    'The following calls were recorded:\n'
                    '\n'
                    '%s'
                    % (
                        self._format_spy_or_call(spy_or_call),
                        safe_repr(expected_args),
                        format_spy_kwargs(expected_kwargs),
                        self._format_spy_calls(
                            spy_or_call,
                            self._format_spy_call_args),
                    ))

    def assertSpyCalledOnceWith(self, spy, *expected_args, **expected_kwargs):
        """Assert that a function was called once with the given arguments.

        This will fail if there was more than one call to the function.

        This will imply :py:meth:`assertSpyCalledOnce`.

        Args:
            spy (callable or kgb.spies.FunctionSpy):
                The function or spy to check.

            *expected_args (tuple):
                Positional arguments expected to be provided in any of the
                calls.

            **expected_kwargs (dict):
                Keyword arguments expected to be provided in any of the calls.

        Raises:
            AssertionError:
                The function was not called with the provided arguments.
        """
        self.assertSpyCalledOnce(spy)

        if not spy.called_with(*expected_args, **expected_kwargs):
            if isinstance(spy, SpyCall):
                self._kgb_assert_fail(
                    'This call to %s was not passed args=%s, kwargs=%s.\n'
                    '\n'
                    'It was called with:\n'
                    '\n'
                    '%s'
                    % (
                        self._format_spy_or_call(spy),
                        safe_repr(expected_args),
                        format_spy_kwargs(expected_kwargs),
                        self._format_spy_call_args(spy),
                    ))
            else:
                self._kgb_assert_fail(
                    'No call to %s was passed args=%s, kwargs=%s.\n'
                    '\n'
                    'The following calls were recorded:\n'
                    '\n'
                    '%s'
                    % (
                        self._format_spy_or_call(spy),
                        safe_repr(expected_args),
                        format_spy_kwargs(expected_kwargs),
                        self._format_spy_calls(
                            spy,
                            self._format_spy_call_args),
                    ))

    def assertSpyNotCalledWith(self, spy_or_call, *expected_args,
                               **expected_kwargs):
        """Assert that a function was not called with the given arguments.

        If a spy is provided, all calls will be checked for a match.

        This will imply :py:meth:`assertHasSpy`.

        Args:
            spy_or_call (callable or kgb.spies.FunctionSpy):
                The function, spy, or call to check.

            *expected_args (tuple):
                Positional arguments not expected to be provided in any of the
                calls.

            **expected_kwargs (dict):
                Keyword arguments not expected to be provided in any of the
                calls.

        Raises:
            AssertionError:
                The function was called with the provided arguments.
        """
        if isinstance(spy_or_call, FunctionSpy):
            self.assertSpyCalled(spy_or_call)

        if spy_or_call.called_with(*expected_args, **expected_kwargs):
            if isinstance(spy_or_call, SpyCall):
                self._kgb_assert_fail(
                    'This call to %s was unexpectedly passed args=%s, '
                    'kwargs=%s.'
                    % (
                        self._format_spy_or_call(spy_or_call),
                        safe_repr(expected_args),
                        format_spy_kwargs(expected_kwargs),
                    ))
            else:
                self._kgb_assert_fail(
                    'A call to %s was unexpectedly passed args=%s, '
                    'kwargs=%s.\n'
                    '\n'
                    'The following calls were recorded:\n'
                    '\n'
                    '%s'
                    % (
                        self._format_spy_or_call(spy_or_call),
                        safe_repr(expected_args),
                        format_spy_kwargs(expected_kwargs),
                        self._format_spy_calls(
                            spy_or_call,
                            self._format_spy_call_args),
                    ))

    def assertSpyLastCalledWith(self, spy, *expected_args, **expected_kwargs):
        """Assert that a function was last called with the given arguments.

        This will imply :py:meth:`assertHasSpy`.

        Args:
            spy (callable or kgb.spies.FunctionSpy):
                The function or spy to check.

            *expected_args (tuple):
                Positional arguments expected to be provided in the last call.

            **expected_kwargs (dict):
                Keyword arguments expected to be provided in the last call.

        Raises:
            AssertionError:
                The function was not called last with the provided arguments.
        """
        self.assertSpyCalled(spy)

        if not spy.last_called_with(*expected_args, **expected_kwargs):
            self._kgb_assert_fail(
                'The last call to %s was not passed args=%s, kwargs=%s.\n'
                '\n'
                'It was last called with:\n'
                '\n'
                '%s'
                % (
                    self._format_spy_or_call(spy),
                    safe_repr(expected_args),
                    format_spy_kwargs(expected_kwargs),
                    self._format_spy_call_args(spy.last_call),
                ))

    def assertSpyReturned(self, spy_or_call, return_value):
        """Assert that a function call returned the given value.

        If a spy is provided, all calls will be checked for a match.

        This will imply :py:meth:`assertHasSpy`.

        Args:
            spy_or_call (callable or kgb.spies.FunctionSpy or
                         kgb.spies.SpyCall):
                The function, spy, or call to check.

            return_value (object or type):
                The value expected to be returned by any of the calls.

        Raises:
            AssertionError:
                The function never returned the provided value.
        """
        if isinstance(spy_or_call, FunctionSpy):
            self.assertSpyCalled(spy_or_call)

        if not spy_or_call.returned(return_value):
            if isinstance(spy_or_call, SpyCall):
                self._kgb_assert_fail(
                    'This call to %s did not return %s.\n'
                    '\n'
                    'It returned:\n'
                    '\n'
                    '%s'
                    % (
                        self._format_spy_or_call(spy_or_call),
                        safe_repr(return_value),
                        self._format_spy_call_returned(spy_or_call),
                    ))
            else:
                self._kgb_assert_fail(
                    'No call to %s returned %s.\n'
                    '\n'
                    'The following values have been returned:\n'
                    '\n'
                    '%s'
                    % (
                        self._format_spy_or_call(spy_or_call),
                        safe_repr(return_value),
                        self._format_spy_calls(
                            spy_or_call,
                            self._format_spy_call_returned),
                    ))

    def assertSpyLastReturned(self, spy, return_value):
        """Assert that the last function call returned the given value.

        This will imply :py:meth:`assertHasSpy`.

        Args:
            spy (callable or kgb.spies.FunctionSpy):
                The function or spy to check.

            return_value (object or type):
                The value expected to be returned by the last call.

        Raises:
            AssertionError:
                The function's last call did not return the provided value.
        """
        self.assertSpyCalled(spy)

        if not spy.last_returned(return_value):
            self._kgb_assert_fail(
                'The last call to %s did not return %s.\n'
                '\n'
                'It last returned:\n'
                '\n'
                '%s'
                % (
                    self._format_spy_or_call(spy),
                    safe_repr(return_value),
                    self._format_spy_call_returned(spy.last_call),
                ))

    def assertSpyRaised(self, spy_or_call, exception_cls):
        """Assert that a function call raised the given exception type.

        If a spy is provided, all calls will be checked for a match.

        This will imply :py:meth:`assertHasSpy`.

        Args:
            spy_or_call (callable or kgb.spies.FunctionSpy or
                         kgb.spies.SpyCall):
                The function, spy, or call to check.

            exception_cls (type):
                The exception type expected to be raised by one of the calls.

        Raises:
            AssertionError:
                The function never raised the provided exception type.
        """
        if isinstance(spy_or_call, FunctionSpy):
            self.assertSpyCalled(spy_or_call)

        if not spy_or_call.raised(exception_cls):
            if isinstance(spy_or_call, SpyCall):
                if spy_or_call.exception is not None:
                    self._kgb_assert_fail(
                        'This call to %s did not raise %s. It raised %s.'
                        % (
                            self._format_spy_or_call(spy_or_call),
                            exception_cls.__name__,
                            self._format_spy_call_raised(spy_or_call),
                        ))
                else:
                    self._kgb_assert_fail(
                        'This call to %s did not raise an exception.'
                        % self._format_spy_or_call(spy_or_call))
            else:
                has_raised = any(
                    call.exception is not None
                    for call in spy_or_call.calls
                )

                if has_raised:
                    self._kgb_assert_fail(
                        'No call to %s raised %s.\n'
                        '\n'
                        'The following exceptions have been raised:\n\n'
                        '%s'
                        % (
                            self._format_spy_or_call(spy_or_call),
                            exception_cls.__name__,
                            self._format_spy_calls(
                                spy_or_call,
                                self._format_spy_call_raised),
                        ))
                else:
                    self._kgb_assert_fail(
                        'No call to %s raised an exception.'
                        % self._format_spy_or_call(spy_or_call))

    def assertSpyLastRaised(self, spy, exception_cls):
        """Assert that the last function call raised the given exception type.

        This will imply :py:meth:`assertHasSpy`.

        Args:
            spy (callable or kgb.spies.FunctionSpy):
                The function or spy to check.

            exception_cls (type):
                The exception type expected to be raised by the last call.

        Raises:
            AssertionError:
                The last function call did not raise the provided exception
                type.
        """
        self.assertSpyCalled(spy)

        if not spy.last_raised(exception_cls):
            if spy.last_call.exception is not None:
                self._kgb_assert_fail(
                    'The last call to %s did not raise %s. It last '
                    'raised %s.'
                    % (
                        self._format_spy_or_call(spy),
                        exception_cls.__name__,
                        self._format_spy_call_raised(spy.last_call),
                    ))
            else:
                self._kgb_assert_fail(
                    'The last call to %s did not raise an exception.'
                    % self._format_spy_or_call(spy))

    def assertSpyRaisedMessage(self, spy_or_call, exception_cls, message):
        """Assert that a function call raised the given exception/message.

        If a spy is provided, all calls will be checked for a match.

        This will imply :py:meth:`assertHasSpy`.

        Args:
            spy_or_call (callable or kgb.spies.FunctionSpy or
                         kgb.spies.SpyCall):
                The function, spy, or call to check.

            exception_cls (type):
                The exception type expected to be raised by one of the calls.

            message (bytes or unicode):
                The expected message in a matching extension.

        Raises:
            AssertionError:
                The function never raised the provided exception type with
                the expected message.
        """
        if isinstance(spy_or_call, FunctionSpy):
            self.assertSpyCalled(spy_or_call)

        if not spy_or_call.raised_with_message(exception_cls, message):
            if isinstance(spy_or_call, SpyCall):
                if spy_or_call.exception is not None:
                    self._kgb_assert_fail(
                        'This call to %s did not raise %s with message %r.\n'
                        '\n'
                        'It raised:\n'
                        '\n'
                        '%s'
                        % (
                            self._format_spy_or_call(spy_or_call),
                            exception_cls.__name__,
                            message,
                            self._format_spy_call_raised_with_message(
                                spy_or_call),
                        ))
                else:
                    self._kgb_assert_fail(
                        'This call to %s did not raise an exception.'
                        % self._format_spy_or_call(spy_or_call))
            else:
                has_raised = any(
                    call.exception is not None
                    for call in spy_or_call.calls
                )

                if has_raised:
                    self._kgb_assert_fail(
                        'No call to %s raised %s with message %r.\n'
                        '\n'
                        'The following exceptions have been raised:\n'
                        '\n'
                        '%s'
                        % (
                            self._format_spy_or_call(spy_or_call),
                            exception_cls.__name__,
                            message,
                            self._format_spy_calls(
                                spy_or_call,
                                self._format_spy_call_raised_with_message),
                        ))
                else:
                    self._kgb_assert_fail(
                        'No call to %s raised an exception.'
                        % self._format_spy_or_call(spy_or_call))

    def assertSpyLastRaisedMessage(self, spy, exception_cls, message):
        """Assert that the function last raised the given exception/message.

        This will imply :py:meth:`assertHasSpy`.

        Args:
            spy (callable or kgb.spies.FunctionSpy):
                The function or spy to check.

            exception_cls (type):
                The exception type expected to be raised by the last call.

            message (bytes or unicode):
                The expected message in the matching extension.

        Raises:
            AssertionError:
                The last function call did not raise the provided exception
                type with the expected message.
        """
        self.assertSpyCalled(spy)

        if not spy.last_raised_with_message(exception_cls, message):
            if spy.last_call.exception is not None:
                self._kgb_assert_fail(
                    'The last call to %s did not raise %s with message %r.\n'
                    '\n'
                    'It last raised:\n'
                    '\n'
                    '%s'
                    % (
                        self._format_spy_or_call(spy),
                        exception_cls.__name__,
                        message,
                        self._format_spy_call_raised_with_message(
                            spy.last_call),
                    ))
            else:
                self._kgb_assert_fail(
                    'The last call to %s did not raise an exception.'
                    % self._format_spy_or_call(spy))

    def _kgb_assert_fail(self, msg):
        """Raise an assertion failure.

        If this class is mixed into a unit test suite, this will call the
        main :py:meth:`unittest.TestCase.fail` method. Otherwise, it will
        simply raise an :py:exc:`AssertionError`.

        Args:
            msg (unicode):
                The assertion message.

        Raises:
            AssertionError:
                The assertion error to raise.
        """
        if hasattr(self, 'fail') and hasattr(self, 'failureException'):
            # This is likely mixed in to a unit test.
            self.fail(msg)
        else:
            raise AssertionError(msg)

    def _format_spy_or_call(self, spy_or_call):
        """Format a spy or call for output in an assertion message.

        Args:
            spy_or_call (callable or kgb.spies.FunctionSpy or
                         kgb.spies.SpyCall):
                The spy or call to format.

        Returns:
            unicode:
            The formatted name of the function.
        """
        if isinstance(spy_or_call, FunctionSpy):
            spy = spy_or_call.orig_func
        elif isinstance(spy_or_call, SpyCall):
            spy = spy_or_call.spy.orig_func
        else:
            spy = spy_or_call

        name = spy.__name__

        if isinstance(name, bytes):
            name = name.decode('utf-8')

        return name

    def _format_spy_calls(self, spy, formatter):
        """Format a list of calls for a spy.

        Args:
            spy (callable or kgb.spies.FunctionSpy):
                The spy to format.

            formatter (callable):
                A formatting function used for each recorded call.

        Returns:
            unicode:
            The formatted output of the calls.
        """
        return '\n\n'.join(
            'Call %d:\n%s' % (i, formatter(call, indent=2))
            for i, call in enumerate(spy.calls)
        )

    def _format_spy_call_args(self, call, indent=0):
        """Format a call's arguments.

        Args:
            call (kgb.spies.SpyCall):
                The call containing arguments to format.

            indent (int, optional):
                The indentation level for any output.

        Returns:
            unicode:
            The formatted output of the arguments for the call.
        """
        return '%s\n%s' % (
            self._format_spy_lines(call.args,
                                   prefix='args=',
                                   indent=indent),
            self._format_spy_lines(call.kwargs,
                                   prefix='kwargs=',
                                   indent=indent),
        )

    def _format_spy_call_returned(self, call, indent=0):
        """Format the return value from a call.

        Args:
            call (kgb.spies.SpyCall):
                The call containing a return value to format.

            indent (int, optional):
                The indentation level for any output.

        Returns:
            unicode:
            The formatted return value from the call.
        """
        return self._format_spy_lines(call.return_value,
                                      indent=indent)

    def _format_spy_call_raised(self, call, indent=0):
        """Format the exception type raised by a call.

        Args:
            call (kgb.spies.SpyCall):
                The call that raised an exception to format.

            indent (int, optional):
                The indentation level for any output.

        Returns:
            unicode:
            The formatted name of the exception raised by a call.
        """
        return self._format_spy_lines(call.exception.__class__.__name__,
                                      indent=indent,
                                      format_data=False)

    def _format_spy_call_raised_with_message(self, call, indent=0):
        """Format the exception type and message raised by a call.

        Args:
            call (kgb.spies.SpyCall):
                The call that raised an exception to format.

            indent (int, optional):
                The indentation level for any output.

        Returns:
            unicode:
            The formatted name of the exception and accompanying message raised
            by a call.
        """
        return '%s\n%s' % (
            self._format_spy_lines(call.exception.__class__.__name__,
                                   prefix='exception=',
                                   indent=indent,
                                   format_data=False),
            self._format_spy_lines(str(call.exception),
                                   prefix='message=',
                                   indent=indent),
        )

    def _format_spy_lines(self, data, prefix='', indent=0, format_data=True):
        """Format a multi-line list of output for an assertion message.

        Unless otherwise specified, the provided data will be formatted
        using :py:func:`pprint.pformat`.

        The first line of data will be prefixed, if a prefix is provided.
        Subsequent lines be aligned with the contents after the prefix.

        All line will be indented by the given amount.

        Args:
            data (object):
                The data to format.

            prefix (unicode, optional):
                An optional prefix for the first line in the data.

            indent (int, optional):
                The indentation level for any output.

            format_data (bool, optional):
                Whether to format the provided ``data`` using
                :py:func:`pprint.pformat`.

        Returns:
            unicode:
            The formatted string for the data.
        """
        indent_str = ' ' * indent

        if format_data:
            data = pformat(data)

        data_lines = data.splitlines()
        lines = ['%s%s%s' % (indent_str, prefix, data_lines[0])]

        if len(data_lines) > 1:
            indent_str = ' ' * (indent + len(prefix))

            lines += [
                '%s%s' % (indent_str, line)
                for line in data_lines[1:]
            ]

        return '\n'.join(lines)

    # snake_case versions of the test functions.
    #
    # Useful for pytest and other uses.
    assert_has_spy = assertHasSpy
    assert_spy_called = assertSpyCalled
    assert_spy_called_once = assertSpyCalledOnce
    assert_spy_not_called = assertSpyNotCalled
    assert_spy_call_count = assertSpyCallCount
    assert_spy_called_with = assertSpyCalledWith
    assert_spy_called_once_with = assertSpyCalledOnceWith
    assert_spy_not_called_with = assertSpyNotCalledWith
    assert_spy_last_called_with = assertSpyLastCalledWith
    assert_spy_returned = assertSpyReturned
    assert_spy_last_returned = assertSpyLastReturned
    assert_spy_raised = assertSpyRaised
    assert_spy_last_raised = assertSpyLastRaised
    assert_spy_raised_message = assertSpyRaisedMessage
    assert_spy_last_raised_message = assertSpyLastRaisedMessage



"""Configures pytest for kgb.

This will conditionally ignore Python 3 test files on Python 2.
"""

from __future__ import unicode_literals

import sys


if sys.version_info[0] < 3:
    collect_ignore_glob = ['kgb/tests/py3/*']



"""Python compatibility functions and types."""

from __future__ import unicode_literals

import sys


pyver = sys.version_info[:2]

if pyver[0] == 2:
    text_type = unicode

    def iterkeys(d):
        return d.iterkeys()

    def iteritems(d):
        return d.iteritems()
else:
    text_type = str

    def iterkeys(d):
        return iter(d.keys())

    def iteritems(d):
        return iter(d.items())



from __future__ import unicode_literals

import re
import sys
import textwrap

if sys.version_info[:2] > (2, 7):
    import unittest
else:
    import unittest2 as unittest

from kgb.agency import SpyAgency


class MathClass(object):
    def do_math(self, a=1, b=2, *args, **kwargs):
        assert isinstance(self, MathClass)

        return a + b

    def do_math_pos(self, a, b):
        assert isinstance(self, MathClass)

        return a + b

    def do_math_mixed(self, a, b=2, *args, **kwargs):
        assert isinstance(self, MathClass)

        return a + b

    @classmethod
    def class_do_math(cls, a=2, b=5, *args, **kwargs):
        assert issubclass(cls, MathClass)

        return a * b

    @classmethod
    def class_do_math_pos(cls, a, b):
        assert issubclass(cls, MathClass)

        return a * b


class TestCase(unittest.TestCase):
    """Base class for test cases for kgb."""

    ws_re = re.compile(r'\s+')

    def setUp(self):
        self.agency = SpyAgency()
        self.orig_class_do_math = MathClass.class_do_math

    def tearDown(self):
        MathClass.class_do_math = self.orig_class_do_math
        self.agency.unspy_all()

    def shortDescription(self):
        """Return the description of the current test.

        This changes the default behavior to replace all newlines with spaces,
        allowing a test description to span lines. It should still be kept
        short, though.

        Returns:
            bytes:
            The description of the test.
        """
        doc = self._testMethodDoc

        if doc is not None:
            doc = doc.split('\n\n', 1)[0]
            doc = self.ws_re.sub(' ', doc).strip()

        return doc

    def make_func(self, code_str, func_name='func'):
        """Return a new function, created by the supplied Python code.

        This is used to create functions with signatures that depend on a
        specific version of Python, and would generate syntax errors on
        earlier versions.

        Args:
            code_str (unicode):
                The Python code used to create the function.

            func_name (unicode, optional):
                The expected name of the function.

        Returns:
            callable:
            The resulting function.

        Raises:
            Exception:
                There was an error with the supplied code.
        """
        scope = {}
        exec(textwrap.dedent(code_str), scope)

        return scope[func_name]



"""Pytest plugin for kgb.

Version Added:
    7.0
"""

from __future__ import unicode_literals

import pytest

from kgb import SpyAgency


@pytest.fixture
def spy_agency():
    """Provide a KGB spy agency to a Pytest unit test.

    Yields:
        kgb.SpyAgency:
        The spy agency.
    """
    agency = SpyAgency()

    try:
        yield agency
    finally:
        agency.unspy_all()
